# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fcntl
import logging
import os
import select
import shutil
import socket
import subprocess
import threading
import time
import shlex
import uuid

from pathlib import Path

from nodepool import config as nodepool_config
from nodepool import exceptions
from nodepool import provider_manager
from nodepool import stats
from nodepool import zk


MINS = 60
HOURS = 60 * MINS
# How long to wait for an image save
IMAGE_TIMEOUT = 6 * HOURS

# How long to wait between checks for ZooKeeper connectivity if it disappears.
SUSPEND_WAIT_TIME = 30

# HP Cloud requires qemu compat with 0.10. That version works elsewhere,
# so just hardcode it for all qcow2 building
DEFAULT_QEMU_IMAGE_COMPAT_OPTIONS = "--qemu-img-options 'compat=0.10'"

# DIB process polling timeout, in milliseconds
BUILD_PROCESS_POLL_TIMEOUT = 30 * 1000


class DibImageFile(object):
    '''
    Class used as an API to finding locally built DIB image files, and
    also used to represent the found files. Image files are named using
    a unique ID, but can be available in multiple formats (with different
    extensions).
    '''
    def __init__(self, image_id, extension=None):
        self.image_id = image_id
        self.extension = extension
        self.md5 = None
        self.md5_file = None
        self.sha256 = None
        self.sha256_file = None

        # File extension is compared to image type (sans '.') so we
        # store it the same way.
        if self.extension:
            self.extension = self.extension.lstrip('.')

    @staticmethod
    def from_path(path):
        if isinstance(path, str):
            path = Path(path)
        return DibImageFile(path.stem, path.suffix)

    @staticmethod
    def from_image_id(images_dir, image_id):
        images = []
        for image_file in Path(images_dir).iterdir():
            if image_file.is_file():
                image = DibImageFile.from_path(image_file)
                if image.image_id == image_id:
                    images.append(image)
        return images

    @staticmethod
    def from_images_dir(images_dir):
        return [DibImageFile.from_path(x) for x in Path(images_dir).iterdir()]

    def to_path(self, images_dir, with_extension=True):
        my_path = Path(images_dir) / self.image_id
        if with_extension:
            if self.extension is None:
                raise exceptions.BuilderError(
                    'Cannot specify image extension of None'
                )
            my_path = my_path.with_suffix('.' + self.extension)

        # Path.with_suffix() will replace an existing suffix, so we create
        # new Path objects from strings for the checksum files.
        md5_path = Path(str(my_path) + '.md5')
        md5 = self._checksum(md5_path)
        if md5:
            self.md5_file = str(md5_path)
            self.md5 = md5[0:32]

        sha256_path = Path(str(my_path) + '.sha256')
        sha256 = self._checksum(sha256_path)
        if sha256:
            self.sha256_file = str(sha256_path)
            self.sha256 = sha256[0:64]

        return str(my_path)

    def _checksum(self, filename):
        if filename.is_file():
            return filename.read_text()
        return None


class BaseWorker(threading.Thread):
    def __init__(self, builder_id, config_path, secure_path, interval, zk):
        super(BaseWorker, self).__init__()
        self.log = logging.getLogger("nodepool.builder.BaseWorker")
        self.daemon = True
        self._running = False
        self._stop_event = threading.Event()
        self._config = None
        self._config_path = config_path
        self._secure_path = secure_path
        self._zk = zk
        self._hostname = socket.getfqdn()
        self._statsd = stats.get_client()
        self._interval = interval
        self._builder_id = builder_id

    def _checkForZooKeeperChanges(self, new_config):
        '''
        Check config for ZooKeeper cluster changes.

        If the defined set of ZooKeeper servers changes, the connection
        will use the new server set.
        '''
        if self._config.zookeeper_servers != new_config.zookeeper_servers:
            self.log.debug("Detected ZooKeeper server changes")
            self._zk.resetHosts(list(new_config.zookeeper_servers.values()))

    @property
    def running(self):
        return self._running

    def shutdown(self):
        self._running = False
        self._stop_event.set()


class CleanupWorker(BaseWorker):
    '''
    The janitor of nodepool-builder that will remove images from providers
    and any local DIB builds.
    '''

    def __init__(self, name, builder_id, config_path, secure_path,
                 interval, zk):
        super(CleanupWorker, self).__init__(builder_id, config_path,
                                            secure_path, interval, zk)
        self.log = logging.getLogger(
            "nodepool.builder.CleanupWorker.%s" % name)
        self.name = 'CleanupWorker.%s' % name

    def _buildUploadRecencyTable(self):
        '''
        Builds a table for each image of the most recent uploads to each
        provider.

        Example)

            image1:
                providerA: [ (build_id, upload_id, upload_time), ...  ]
                providerB: [ (build_id, upload_id, upload_time), ...  ]
            image2:
                providerC: [ (build_id, upload_id, upload_time), ...  ]
        '''
        self._rtable = {}
        for image in self._zk.getImageNames():
            self._rtable[image] = {}
            for build in self._zk.getBuilds(image, zk.READY):
                for provider in self._zk.getBuildProviders(image, build.id):
                    if provider not in self._rtable[image]:
                        self._rtable[image][provider] = []
                    uploads = self._zk.getMostRecentBuildImageUploads(
                        2, image, build.id, provider, zk.READY)
                    for upload in uploads:
                        self._rtable[image][provider].append(
                            (build.id, upload.id, upload.state_time)
                        )

        # Sort uploads by state_time (upload time) and keep the 2 most recent
        for i in list(self._rtable.keys()):
            for p in self._rtable[i].keys():
                self._rtable[i][p].sort(key=lambda x: x[2], reverse=True)
                self._rtable[i][p] = self._rtable[i][p][:2]

    def _isRecentUpload(self, image, provider, build_id, upload_id):
        '''
        Search for an upload for a build within the recency table.
        '''
        provider = self._rtable[image].get(provider)
        if not provider:
            return False

        for b_id, u_id, u_time in provider:
            if build_id == b_id and upload_id == u_id:
                return True
        return False

    def _inProgressUpload(self, upload):
        '''
        Determine if an upload is in progress.
        '''
        if upload.state != zk.UPLOADING:
            return False

        try:
            with self._zk.imageUploadLock(upload.image_name, upload.build_id,
                                          upload.provider_name,
                                          blocking=False):
                pass
        except exceptions.ZKLockException:
            return True
        return False

    @staticmethod
    def deleteLocalBuild(images_dir, image, build, log):
        '''
        Remove expired image build from local disk.

        It is safe to call this multiple times, or if no local files exist.

        :param str images_dir: Path to the DIB images directory.
        :param str image: Name of the image whose build we are deleting.
        :param ImageBuild build: The build we want to delete.
        :param Logger log: A logging object for log output.
        '''
        def removeDibItem(filename):
            if filename is None:
                return
            try:
                os.remove(filename)
                log.info("Removed DIB file %s" % filename)
            except OSError as e:
                if e.errno != 2:    # No such file or directory
                    raise e

        base = "-".join([image, build.id])
        files = DibImageFile.from_image_id(images_dir, base)
        if not files:
            return

        log.info("Doing cleanup for %s:%s" % (image, build.id))

        manifest_dir = None

        for f in files:
            filename = f.to_path(images_dir, True)
            if not manifest_dir:
                path, ext = filename.rsplit('.', 1)
                manifest_dir = path + ".d"
            items = [filename, f.md5_file, f.sha256_file]
            list(map(removeDibItem, items))

        try:
            shutil.rmtree(manifest_dir)
            log.info("Removed DIB manifest %s" % manifest_dir)
        except OSError as e:
            if e.errno != 2:    # No such file or directory
                raise e

    def _deleteLocalBuild(self, image, build):
        CleanupWorker.deleteLocalBuild(
            self._config.imagesdir, image, build, self.log)

    def _cleanupProvider(self, provider, image, build_id):
        all_uploads = self._zk.getUploads(image, build_id, provider.name)

        for upload in all_uploads:
            if self._isRecentUpload(image, provider.name, build_id, upload.id):
                continue
            self.log.debug("Deleting non-recent image upload: %s", upload)
            self._deleteUpload(upload)

    def _cleanupObsoleteProviderUploads(self, provider, image, build_id):
        if image in provider.diskimages:
            # This image is in use for this provider
            return

        all_uploads = self._zk.getUploads(image, build_id, provider.name)
        for upload in all_uploads:
            self.log.debug("Deleting obsolete image upload: %s", upload)
            self._deleteUpload(upload)

    def _deleteUpload(self, upload):
        if upload.state != zk.DELETING:
            if not self._inProgressUpload(upload):
                try:
                    with self._zk.imageUploadNumberLock(upload,
                                                        blocking=False):
                        upload.state = zk.DELETING
                        self._zk.storeImageUpload(upload.image_name,
                                                  upload.build_id,
                                                  upload.provider_name,
                                                  upload,
                                                  upload.id)
                except exceptions.ZKLockException:
                    # If we can't get a lock, we'll try again later.
                    self.log.debug("Unable to get lock on image upload: %s",
                                   upload)
                    return

        if upload.state == zk.DELETING:
            manager = self._config.provider_managers[upload.provider_name]
            try:
                # It is possible we got this far, but don't actually have an
                # external_id. This could mean that zookeeper and cloud
                # provider are some how out of sync.
                if upload.external_id:
                    base = "-".join([upload.image_name, upload.build_id])
                    self.log.info("Deleting image build %s from %s" %
                                  (base, upload.provider_name))
                    manager.deleteImage(upload.external_name,
                                        upload.external_id)
            except Exception:
                self.log.exception(
                    "Unable to delete image %s from %s:",
                    upload.external_name, upload.provider_name)
            else:
                self.log.debug("Deleting image upload: %s", upload)
                try:
                    with self._zk.imageUploadNumberLock(upload,
                                                        blocking=False):
                        self._zk.deleteUpload(upload.image_name,
                                              upload.build_id,
                                              upload.provider_name, upload.id)
                except exceptions.ZKLockException:
                    # If we can't get a lock, we'll try again later.
                    self.log.debug("Unable to get lock on image upload: %s",
                                   upload)

    def _inProgressBuild(self, build, image):
        '''
        Determine if a DIB build is in progress.
        '''
        if build.state != zk.BUILDING:
            return False

        try:
            with self._zk.imageBuildLock(image, blocking=False):
                # An additional state check is needed to make sure it hasn't
                # changed on us. If it has, then let's pretend a build is
                # still in progress so that it is checked again later with
                # its new build state.
                b = self._zk.getBuild(image, build.id)
                if b.state != zk.BUILDING:
                    return True
                pass
        except exceptions.ZKLockException:
            return True
        return False

    def _cleanup(self):
        '''
        Clean up builds on disk and in providers.
        '''
        known_providers = self._config.providers.values()
        image_names = self._zk.getImageNames()

        self._buildUploadRecencyTable()

        for image in image_names:
            try:
                self._cleanupImage(known_providers, image)
            except Exception:
                self.log.exception("Exception cleaning up image %s:", image)

    def _filterLocalBuilds(self, image, builds):
        '''Return the subset of builds that are local'''
        ret = []
        for build in builds:
            is_local = build.builder_id == self._builder_id
            if is_local:
                ret.append(build)
        return ret

    def _cleanupCurrentProviderUploads(self, provider, image, build_id):
        '''
        Remove cruft from a current build.

        Current builds (the ones we want to keep) are treated special since
        we want to remove any ZK nodes for uploads that failed exceptionally
        hard (i.e., we could not set the state to FAILED and they remain as
        UPLOADING), and we also want to remove any uploads that have been
        marked for deleting.
        '''
        cruft = self._zk.getUploads(image, build_id, provider,
                                    states=[zk.UPLOADING,
                                            zk.DELETING,
                                            zk.FAILED])
        for upload in cruft:
            if (upload.state == zk.UPLOADING and
                not self._inProgressUpload(upload)
            ):
                # Since we cache the uploads above, we need to verify the
                # state hasn't changed on us (e.g., it could have gone from
                # an in progress upload to a successfully completed upload
                # between the getUploads() and the _inProgressUpload() check.
                u = self._zk.getImageUpload(image, build_id, provider,
                                            upload.id)
                if upload.state != u.state:
                    continue
                self.log.debug("Removing failed upload record: %s" % upload)
                self._zk.deleteUpload(image, build_id, provider, upload.id)
            elif upload.state == zk.DELETING:
                self.log.debug(
                    "Removing deleted upload and record: %s" % upload)
                self._deleteUpload(upload)
            elif upload.state == zk.FAILED:
                self.log.debug(
                    "Removing failed upload and record: %s" % upload)
                self._deleteUpload(upload)

    def _cleanupImage(self, known_providers, image):
        '''
        Clean up one image.
        '''
        # Get the list of all builds, then work from that so that we
        # have a consistent view of the data.
        all_builds = self._zk.getBuilds(image)
        builds_to_keep = set([b for b in sorted(all_builds, reverse=True,
                                                key=lambda y: y.state_time)
                              if b.state == zk.READY][:2])
        local_builds = set(self._filterLocalBuilds(image, all_builds))
        diskimage = self._config.diskimages.get(image)
        if not diskimage and not local_builds:
            # This builder is not configured to build this image and was not
            # responsible for this image build so ignore it.
            return
        # Remove any local builds that are not in use.
        if not diskimage or (diskimage and not diskimage.image_types):
            builds_to_keep -= local_builds
            # TODO(jeblair): When all builds for an image which is not
            # in use are deleted, the image znode should be deleted as
            # well.

        for build in all_builds:
            # Start by deleting any uploads that are no longer needed
            # because this image has been removed from a provider
            # (since this should be done regardless of the build
            # state).
            for provider in known_providers:
                if not provider.manage_images:
                    # This provider doesn't manage images
                    continue
                try:
                    self._cleanupObsoleteProviderUploads(provider, image,
                                                         build.id)
                    if build in builds_to_keep:
                        self._cleanupCurrentProviderUploads(provider.name,
                                                            image,
                                                            build.id)
                except Exception:
                    self.log.exception("Exception cleaning up uploads "
                                       "of build %s of image %s in "
                                       "provider %s:",
                                       build, image, provider)

            # If the build is in the delete state, we will try to
            # delete the entire thing regardless.
            if build.state != zk.DELETING:
                # If it is in any other state, we will only delete it
                # if it is older than the most recent two ready
                # builds, or is in the building state but not actually
                # building.
                if build in builds_to_keep:
                    continue
                elif self._inProgressBuild(build, image):
                    continue

            for provider in known_providers:
                if not provider.manage_images:
                    # This provider doesn't manage images
                    continue
                try:
                    self._cleanupProvider(provider, image, build.id)
                except Exception:
                    self.log.exception("Exception cleaning up build %s "
                                       "of image %s in provider %s:",
                                       build, image, provider)

            uploads = []
            for p in self._zk.getBuildProviders(image, build.id):
                uploads += self._zk.getUploads(image, build.id, p)

            # If we own this build, delete the local DIB files as soon as all
            # provider uploads are in a deleting state. This prevents us from
            # keeping local files around while we wait on clouds to remove
            # the image on their side (which can be very slow).
            all_deleting = all(map(lambda x: x.state == zk.DELETING, uploads))
            is_local = build.builder_id == self._builder_id
            if (not uploads or all_deleting) and is_local:
                # If we got here, it's either an explicit delete (user
                # activated), or we're deleting because we have newer
                # images.  We're about to start deleting files, so
                # make sure that the dib image state reflects that.
                if build.state != zk.DELETING:
                    with self._zk.imageBuildNumberLock(
                        image, build.id, blocking=False
                    ):
                        build.state = zk.DELETING
                        self._zk.storeBuild(image, build, build.id)

                self._deleteLocalBuild(image, build)

            if not uploads:
                if is_local:
                    # Only remove the db build record from the builder that
                    # built this image. This prevents us from removing the db
                    # record without first removing the local image files.
                    if not self._zk.deleteBuild(image, build.id):
                        self.log.error("Unable to delete build %s because"
                                       " uploads still remain.", build)

    def run(self):
        '''
        Start point for the CleanupWorker thread.
        '''
        self._running = True
        while self._running:
            # Don't do work if we've lost communication with the ZK cluster
            did_suspend = False
            while self._zk and (self._zk.suspended or self._zk.lost):
                did_suspend = True
                self.log.info("ZooKeeper suspended. Waiting")
                time.sleep(SUSPEND_WAIT_TIME)
            if did_suspend:
                self.log.info("ZooKeeper available. Resuming")

            try:
                self._run()
            except Exception:
                self.log.exception("Exception in CleanupWorker:")
                time.sleep(10)

            self._stop_event.wait(self._interval)

        provider_manager.ProviderManager.stopProviders(self._config)

    def _run(self):
        '''
        Body of run method for exception handling purposes.
        '''
        new_config = nodepool_config.loadConfig(self._config_path)
        if self._secure_path:
            nodepool_config.loadSecureConfig(new_config, self._secure_path)
        if not self._config:
            self._config = new_config

        self._checkForZooKeeperChanges(new_config)
        provider_manager.ProviderManager.reconfigure(self._config, new_config,
                                                     self._zk,
                                                     only_image_manager=True)
        self._config = new_config

        self._cleanup()


class BuildWorker(BaseWorker):
    def __init__(self, name, builder_id, config_path, secure_path,
                 interval, zk):
        super(BuildWorker, self).__init__(builder_id, config_path, secure_path,
                                          interval, zk)
        self.log = logging.getLogger("nodepool.builder.BuildWorker.%s" % name)
        self.name = 'BuildWorker.%s' % name

    def _getBuildLogRoot(self, name):
        log_dir = self._config.build_log_dir
        if not log_dir:
            log_dir = '/var/log/nodepool/builds'
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
        return log_dir

    def _pruneBuildLogs(self, name):
        if self._config.build_log_retention < 0:
            return
        log_dir = self._getBuildLogRoot(name)
        keep = max(self._config.build_log_retention, 1)
        existing = sorted(os.listdir(log_dir))
        existing = [f for f in existing if f.startswith(name)]
        delete = existing[:0 - keep]
        for f in delete:
            fp = os.path.join(log_dir, f)
            self.log.info("Deleting old build log %s" % (fp,))
            os.unlink(fp)

    def _getBuildLog(self, name, build_id):
        log_dir = self._getBuildLogRoot(name)
        return os.path.join(log_dir, '%s-%s.log' % (name, build_id))

    def _checkForScheduledImageUpdates(self):
        '''
        Check every DIB image to see if it has aged out and needs rebuilt.
        '''
        for diskimage in self._config.diskimages.values():
            # Check if we've been told to shutdown
            # or if ZK connection is suspended
            if not self._running or self._zk.suspended or self._zk.lost:
                return
            try:
                self._checkImageForScheduledImageUpdates(diskimage)
            except Exception:
                self.log.exception("Exception checking for scheduled "
                                   "update of diskimage %s",
                                   diskimage.name)

    def _checkImageForScheduledImageUpdates(self, diskimage):
        '''
        Check one DIB image to see if it needs to be rebuilt.

        .. note:: It's important to lock the image build before we check
            the state time and then build to eliminate any race condition.
        '''
        # Check if diskimage builds are paused.
        if diskimage.pause:
            return

        if not diskimage.image_types:
            # We don't know what formats to build.
            return

        now = int(time.time())
        builds = self._zk.getMostRecentBuilds(1, diskimage.name, zk.READY)

        # If there is no build for this image, or it has aged out
        # or if the current build is missing an image type from
        # the config file, start a new build.
        if (not builds
            or (now - builds[0].state_time) >= diskimage.rebuild_age
            or not set(builds[0].formats).issuperset(diskimage.image_types)
            ):

            try:
                with self._zk.imageBuildLock(diskimage.name, blocking=False):
                    # To avoid locking each image repeatedly, we have an
                    # second, redundant check here to verify that a new
                    # build didn't appear between the first check and the
                    # lock acquisition. If it's not the same build as
                    # identified in the first check above, assume another
                    # BuildWorker created the build for us and continue.
                    builds2 = self._zk.getMostRecentBuilds(
                        1, diskimage.name, zk.READY)
                    if builds2 and builds[0].id != builds2[0].id:
                        return

                    self.log.info("Building image %s" % diskimage.name)
                    self._buildWrapper(diskimage)
            except exceptions.ZKLockException:
                # Lock is already held. Skip it.
                pass

    def _buildWrapper(self, diskimage):
        '''
        Wraps logic for disk image building and ZooKeeper recording.

        :returns: The updated ImageBuild data structure.
        '''
        data = zk.ImageBuild()
        data.state = zk.BUILDING
        data.builder_id = self._builder_id
        data.builder = self._hostname
        data.formats = list(diskimage.image_types)

        bnum = self._zk.storeBuild(diskimage.name, data)

        try:
            data = self._buildImage(bnum, diskimage)
        except Exception:
            # If something unexpected happens, make sure we clean up any
            # build cruft by executing the fallthrough below.
            self.log.exception("Image build failure for build %s of image %s:",
                               bnum, diskimage.name)
            data.state = zk.FAILED

        # If we lost the session during the long running build, this means
        # we've lost our lock and the node was possibly cleaned up during the
        # long build time because state was BUILDING and unlocked. And because
        # the cleanup likely did not get the in-progress dib build files, we
        # need to make sure we get the rest here.
        try:
            self._zk.storeBuild(diskimage.name, data, bnum)
        except Exception:
            # If it disappeared, or the save otherwise failed, cleanup the
            # leaked build manually.
            self.log.exception(
                "Unable to update record for build %s of image %s:",
                bnum, diskimage.name)
            data.id = bnum
            CleanupWorker.deleteLocalBuild(
                self._config.imagesdir, diskimage.name, data, self.log)
            data.state = zk.FAILED
            return data

        return self._zk.getBuild(diskimage.name, bnum)

    def _checkForManualBuildRequest(self):
        '''
        Query ZooKeeper for any manual image build requests.
        '''
        for diskimage in self._config.diskimages.values():
            # Check if we've been told to shutdown
            # or if ZK connection is suspended
            if not self._running or self._zk.suspended or self._zk.lost:
                return
            try:
                self._checkImageForManualBuildRequest(diskimage)
            except Exception:
                self.log.exception("Exception checking for manual "
                                   "update of diskimage %s",
                                   diskimage)

    def _checkImageForManualBuildRequest(self, diskimage):
        '''
        Query ZooKeeper for a manual image build request for one image.
        '''
        # Check if diskimage builds are paused.
        if diskimage.pause:
            return

        # Reduce use of locks by adding an initial check here and
        # a redundant check after lock acquisition.
        if not self._zk.hasBuildRequest(diskimage.name):
            return

        try:
            with self._zk.imageBuildLock(diskimage.name, blocking=False):
                # Redundant check
                if not self._zk.hasBuildRequest(diskimage.name):
                    return

                self.log.info(
                    "Manual build request for image %s" % diskimage.name)
                data = self._buildWrapper(diskimage)

                # Remove request on a successful build
                if data.state == zk.READY:
                    self._zk.removeBuildRequest(diskimage.name)

        except exceptions.ZKLockException:
            # Lock is already held. Skip it.
            pass

    def _buildImage(self, build_id, diskimage):
        '''
        Run the external command to build the diskimage.

        :param str build_id: The ID for the build (used in image filename).
        :param diskimage: The diskimage as retrieved from our config file.

        :returns: An ImageBuild object of build-related data.

        :raises: BuilderError if we failed to execute the build command.
        '''
        base = "-".join([diskimage.name, build_id])
        image_file = DibImageFile(base)
        filename = image_file.to_path(self._config.imagesdir, False)

        env = os.environ.copy()
        env['DIB_RELEASE'] = diskimage.release
        env['DIB_IMAGE_NAME'] = diskimage.name
        env['DIB_IMAGE_FILENAME'] = filename

        # Note we use a reference to the nodepool config here so
        # that whenever the config is updated we get up to date
        # values in this thread.
        if self._config.elementsdir:
            env['ELEMENTS_PATH'] = self._config.elementsdir

        # send additional env vars if needed
        for k, v in diskimage.env_vars.items():
            env[k] = v

        img_elements = diskimage.elements
        img_types = ",".join(diskimage.image_types)

        qemu_img_options = ''
        if 'qcow2' in img_types:
            qemu_img_options = DEFAULT_QEMU_IMAGE_COMPAT_OPTIONS

        # A bit of a hack, but useful for CI to pick up the
        # fake-image-create relative to this file easily
        dib_cmd = diskimage.dib_cmd.replace("%p", os.path.dirname(__file__))

        cmd = ('%s -x -t %s --checksum --no-tmpfs %s -o %s %s' %
               (dib_cmd, img_types, qemu_img_options,
                filename, img_elements))

        self._pruneBuildLogs(diskimage.name)
        log_fn = self._getBuildLog(diskimage.name, build_id)

        self.log.info('Running %s' % (cmd,))
        self.log.info('Logging to %s' % (log_fn,))

        start_time = time.monotonic()

        try:
            p = subprocess.Popen(
                shlex.split(cmd),
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                env=env)
        except OSError as e:
            raise exceptions.BuilderError(
                "Failed to exec '%s'. Error: '%s'" % (cmd, e.strerror)
            )

        # Make subprocess stdout non-blocking
        fd = p.stdout.fileno()
        fl = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)

        # Poll subprocess stdout for readability
        r, w, e = select.select([fd], [], [], 0)
        bitmask = (select.POLLIN | select.POLLHUP)
        poll = select.poll()
        poll.register(fd, bitmask)

        rc = None
        did_timeout = False
        subprocess_done = False

        def buildDidTimeout():
            build_time = time.monotonic() - start_time
            if build_time > diskimage.build_timeout:
                return True
            return False

        with open(log_fn, 'wb') as log:

            # While the subprocess is running, we will loop through stdout
            # events. If we can read data, write that out to the log file.
            # If the subprocess completes, set the flag so we can exit and
            # write the return code.
            #
            # We check for build timeouts in two places: after we have read
            # all available data in the buffer (which will cause an early exit
            # of the poll loop), and after a poll timeout. If we did not have
            # a check after the reads, we would have to have a poll timeout
            # to occur to check for a build timeout, which may not happen if
            # there is constantly data available for reading.

            while not subprocess_done:
                for fd, event in poll.poll(BUILD_PROCESS_POLL_TIMEOUT):

                    # Data available for reading
                    if event & select.POLLIN:
                        data = p.stdout.read(1024)
                        while data:
                            log.write(data)
                            log.flush()
                            data = p.stdout.read(1024)
                        if buildDidTimeout():
                            break

                    # Subprocess complete
                    elif event & select.POLLHUP:
                        subprocess_done = True
                        rc = p.wait()

                if not subprocess_done:
                    if buildDidTimeout():
                        did_timeout = True
                        rc = 1
                        self.log.error(
                            "Build timeout for image %s, build %s (log: %s)",
                            diskimage.name, build_id, log_fn)
                        p.kill()
                        break

            # Subprocess finished, write return code
            if not did_timeout:
                m = "Exit code: %s\n" % rc
                log.write(m.encode('utf8'))

        # It's possible the connection to the ZK cluster could have been
        # interrupted during the build. If so, wait for it to return.
        # It could transition directly from SUSPENDED to CONNECTED, or go
        # through the LOST state before CONNECTED.
        did_suspend = False
        while self._zk.suspended or self._zk.lost:
            did_suspend = True
            self.log.info("ZooKeeper suspended during build. Waiting")
            time.sleep(SUSPEND_WAIT_TIME)
        if did_suspend:
            self.log.info("ZooKeeper available. Resuming")

        build_time = time.monotonic() - start_time

        build_data = zk.ImageBuild()
        build_data.builder_id = self._builder_id
        build_data.builder = self._hostname
        build_data.username = diskimage.username
        build_data.python_path = diskimage.python_path

        if self._statsd:
            pipeline = self._statsd.pipeline()

        if self._zk.didLoseConnection:
            self.log.info("ZooKeeper lost while building %s" % diskimage.name)
            self._zk.resetLostFlag()
            build_data.state = zk.FAILED
        elif p.returncode or did_timeout:
            self.log.info(
                "DIB failed creating %s (%s) (timeout=%s)" % (
                    diskimage.name, p.returncode, did_timeout))
            build_data.state = zk.FAILED
        else:
            self.log.info("DIB image %s is built" % diskimage.name)
            build_data.state = zk.READY
            build_data.formats = list(diskimage.image_types)

            if self._statsd:
                # record stats on the size of each image we create
                for ext in img_types.split(','):
                    key = 'nodepool.dib_image_build.%s.%s.size' % (
                        diskimage.name, ext)
                    # A bit tricky because these image files may be sparse
                    # files; we only want the true size of the file for
                    # purposes of watching if we've added too much stuff
                    # into the image.  Note that st_blocks is defined as
                    # 512-byte blocks by stat(2)
                    size = os.stat("%s.%s" % (filename, ext)).st_blocks * 512
                    self.log.debug("%s created image %s.%s (size: %d) " %
                                   (diskimage.name, filename, ext, size))
                    pipeline.gauge(key, size)

        if self._statsd:
            # report result to statsd
            key_base = 'nodepool.dib_image_build.%s.status' % (
                diskimage.name)
            pipeline.timing(key_base + '.duration',
                            int(build_time * 1000))
            pipeline.gauge(key_base + '.rc', rc)
            pipeline.gauge(key_base + '.last_build',
                           int(time.time()))
            pipeline.send()

        return build_data

    def run(self):
        '''
        Start point for the BuildWorker thread.
        '''
        self._running = True
        while self._running:
            # Don't do work if we've lost communication with the ZK cluster
            did_suspend = False
            while self._zk and (self._zk.suspended or self._zk.lost):
                did_suspend = True
                self.log.info("ZooKeeper suspended. Waiting")
                time.sleep(SUSPEND_WAIT_TIME)
            if did_suspend:
                self.log.info("ZooKeeper available. Resuming")

            try:
                self._run()
            except Exception:
                self.log.exception("Exception in BuildWorker:")
                time.sleep(10)

            self._stop_event.wait(self._interval)

    def _run(self):
        '''
        Body of run method for exception handling purposes.
        '''
        # NOTE: For the first iteration, we expect self._config to be None
        new_config = nodepool_config.loadConfig(self._config_path)
        if self._secure_path:
            nodepool_config.loadSecureConfig(new_config, self._secure_path)
        if not self._config:
            self._config = new_config

        self._checkForZooKeeperChanges(new_config)
        self._config = new_config
        self._checkForScheduledImageUpdates()
        self._checkForManualBuildRequest()


class UploadWorker(BaseWorker):
    def __init__(self, name, builder_id, config_path, secure_path,
                 interval, zk):
        super(UploadWorker, self).__init__(builder_id, config_path,
                                           secure_path, interval, zk)
        self.log = logging.getLogger("nodepool.builder.UploadWorker.%s" % name)
        self.name = 'UploadWorker.%s' % name

    def _reloadConfig(self):
        '''
        Reload the nodepool configuration file.
        '''
        new_config = nodepool_config.loadConfig(self._config_path)
        if self._secure_path:
            nodepool_config.loadSecureConfig(new_config, self._secure_path)
        if not self._config:
            self._config = new_config

        self._checkForZooKeeperChanges(new_config)
        provider_manager.ProviderManager.reconfigure(self._config, new_config,
                                                     self._zk,
                                                     only_image_manager=True)
        self._config = new_config

    def _uploadImage(self, build_id, upload_id, image_name, images, provider,
                     username, python_path):
        '''
        Upload a local DIB image build to a provider.

        :param str build_id: Unique ID of the image build to upload.
        :param str upload_id: Unique ID of the upload.
        :param str image_name: Name of the diskimage.
        :param list images: A list of DibImageFile objects from this build
            that available for uploading.
        :param provider: The provider from the parsed config file.
        :param username:
        :param python_path:
        '''
        start_time = time.time()
        timestamp = int(start_time)

        image = None
        for i in images:
            if provider.image_type == i.extension:
                image = i
                break

        if not image:
            raise exceptions.BuilderInvalidCommandError(
                "Unable to find image file of type %s for id %s to upload" %
                (provider.image_type, build_id)
            )

        self.log.debug("Found image file of type %s for image id: %s" %
                       (image.extension, image.image_id))

        filename = image.to_path(self._config.imagesdir, with_extension=True)

        ext_image_name = provider.image_name_format.format(
            image_name=image_name, timestamp=str(timestamp)
        )

        self.log.info("Uploading DIB image build %s from %s to %s" %
                      (build_id, filename, provider.name))

        manager = self._config.provider_managers[provider.name]
        provider_image = provider.diskimages.get(image_name)
        if provider_image is None:
            raise exceptions.BuilderInvalidCommandError(
                "Could not find matching provider image for %s" % image_name
            )

        meta = provider_image.meta.copy()
        meta['nodepool_build_id'] = build_id
        meta['nodepool_upload_id'] = upload_id

        try:
            external_id = manager.uploadImage(
                ext_image_name, filename,
                image_type=image.extension,
                meta=meta,
                md5=image.md5,
                sha256=image.sha256,
            )
        except Exception:
            self.log.exception(
                "Failed to upload build %s of image %s to provider %s" %
                (build_id, image_name, provider.name))
            data = zk.ImageUpload()
            data.state = zk.FAILED
            return data

        if provider.post_upload_hook:
            try:
                cmd = [
                    provider.post_upload_hook,
                    provider.name,
                    external_id,
                    filename
                ]
                self.log.info('Running post upload hook %s', cmd)
                p = subprocess.run(cmd, stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE, check=True)
            except Exception as e:
                if isinstance(e, subprocess.CalledProcessError):
                    self.log.error('Post upload hook failed with '
                                   'exit code %s\nstdout:\n%s\nstderr:\n%s',
                                   e.returncode, e.stdout.decode(),
                                   e.stderr.decode())
                else:
                    self.log.exception('Unknown exception during upload hook')

                try:
                    manager.deleteImage(ext_image_name)
                except Exception:
                    # Image delete failed but we cannot do anything about this
                    # right now so just log the exception.
                    self.log.exception('Unable to delete image "%s"',
                                       ext_image_name)
                data = zk.ImageUpload()
                data.state = zk.FAILED
                return data

            self.log.info(
                'Post upload hook success with exit code %s\n'
                'stdout:\n%s\nstderr:\n%s',
                p.returncode, p.stdout.decode(), p.stderr.decode())

        if self._statsd:
            dt = int((time.time() - start_time) * 1000)
            key = 'nodepool.image_update.%s.%s' % (image_name,
                                                   provider.name)
            self._statsd.timing(key, dt)
            self._statsd.incr(key)

        base = "-".join([image_name, build_id])
        self.log.info("Image build %s (external_id %s) in %s is ready" %
                      (base, external_id, provider.name))

        data = zk.ImageUpload()
        data.state = zk.READY
        data.external_id = external_id
        data.external_name = ext_image_name
        data.format = image.extension
        data.username = username
        data.python_path = python_path

        return data

    def _checkForProviderUploads(self):
        '''
        Check for any image builds that need to be uploaded to providers.

        If we find any builds in the 'ready' state that haven't been uploaded
        to providers, do the upload if they are available on the local disk.
        '''
        for provider in self._config.providers.values():
            if not provider.manage_images:
                continue
            for image in provider.diskimages.values():
                uploaded = False

                # Check if we've been told to shutdown
                # or if ZK connection is suspended
                if not self._running or self._zk.suspended or self._zk.lost:
                    return
                try:
                    uploaded = self._checkProviderImageUpload(provider, image)
                except Exception:
                    self.log.exception("Error uploading image %s "
                                       "to provider %s:",
                                       image.name, provider.name)

                # NOTE: Due to the configuration file disagreement issue
                # (the copy we have may not be current), if we took the time
                # to attempt to upload an image, let's short-circuit this loop
                # to give us a chance to reload the configuration file.
                if uploaded:
                    return

    def _checkProviderImageUpload(self, provider, image):
        '''
        The main body of _checkForProviderUploads.  This encapsulates
        checking whether an image for a provider should be uploaded
        and performing the upload.  It is a separate function so that
        exception handling can treat all provider-image uploads
        indepedently.

        :returns: True if an upload was attempted, False otherwise.
        '''
        # Check if image uploads are paused.
        if provider.diskimages.get(image.name).pause:
            return False

        # Search for the most recent 'ready' image build
        builds = self._zk.getMostRecentBuilds(1, image.name,
                                              zk.READY)
        if not builds:
            return False

        build = builds[0]

        # Search for locally built images. The image name and build
        # sequence ID is used to name the image.
        local_images = DibImageFile.from_image_id(
            self._config.imagesdir, "-".join([image.name, build.id]))
        if not local_images:
            return False

        # See if this image has already been uploaded
        upload = self._zk.getMostRecentBuildImageUploads(
            1, image.name, build.id, provider.name, zk.READY)
        if upload:
            return False

        # See if this provider supports the available image formats
        if provider.image_type not in build.formats:
            return False

        try:
            with self._zk.imageUploadLock(
                image.name, build.id, provider.name,
                blocking=False
            ):
                # Verify once more that it hasn't been uploaded since the
                # last check.
                upload = self._zk.getMostRecentBuildImageUploads(
                    1, image.name, build.id, provider.name, zk.READY)
                if upload:
                    return False

                # NOTE: Due to the configuration file disagreement issue
                # (the copy we have may not be current), we try to verify
                # that another thread isn't trying to delete this build just
                # before we upload.
                b = self._zk.getBuild(image.name, build.id)
                if b.state == zk.DELETING:
                    return False

                # New upload number with initial state 'uploading'
                data = zk.ImageUpload()
                data.state = zk.UPLOADING
                data.username = build.username
                data.python_path = build.python_path

                upnum = self._zk.storeImageUpload(
                    image.name, build.id, provider.name, data)

                data = self._uploadImage(build.id, upnum, image.name,
                                         local_images, provider,
                                         build.username, build.python_path)

                # Set final state
                self._zk.storeImageUpload(image.name, build.id,
                                          provider.name, data, upnum)
                return True
        except exceptions.ZKLockException:
            # Lock is already held. Skip it.
            return False

    def run(self):

        '''
        Start point for the UploadWorker thread.
        '''
        self._running = True
        while self._running:
            # Don't do work if we've lost communication with the ZK cluster
            did_suspend = False
            while self._zk and (self._zk.suspended or self._zk.lost):
                did_suspend = True
                self.log.info("ZooKeeper suspended. Waiting")
                time.sleep(SUSPEND_WAIT_TIME)
            if did_suspend:
                self.log.info("ZooKeeper available. Resuming")

            try:
                self._reloadConfig()
                self._checkForProviderUploads()
            except Exception:
                self.log.exception("Exception in UploadWorker:")
                time.sleep(10)

            self._stop_event.wait(self._interval)

        provider_manager.ProviderManager.stopProviders(self._config)


class NodePoolBuilder(object):
    '''
    Main class for the Nodepool Builder.

    The builder has the responsibility to:

        * Start and maintain the working state of each worker thread.
    '''
    log = logging.getLogger("nodepool.builder.NodePoolBuilder")

    def __init__(self, config_path, secure_path=None,
                 num_builders=1, num_uploaders=4):
        '''
        Initialize the NodePoolBuilder object.

        :param str config_path: Path to configuration file.
        :param str secure_path: Path to secure configuration file.
        :param int num_builders: Number of build workers to start.
        :param int num_uploaders: Number of upload workers to start.
        '''
        self._config_path = config_path
        self._secure_path = secure_path
        self._config = None
        self._num_builders = num_builders
        self._build_workers = []
        self._num_uploaders = num_uploaders
        self._upload_workers = []
        self._janitor = None
        self._running = False
        self.cleanup_interval = 60
        self.build_interval = 10
        self.upload_interval = 10
        self.zk = None

        # This lock is needed because the run() method is started in a
        # separate thread of control, which can return before the scheduler
        # has completed startup. We need to avoid shutting down before the
        # startup process has completed.
        self._start_lock = threading.Lock()

    # ======================================================================
    # Private methods
    # ======================================================================

    def _getBuilderID(self, id_file):
        if not os.path.exists(id_file):
            with open(id_file, "w") as f:
                builder_id = str(uuid.uuid4())
                f.write(builder_id)
            return builder_id

        with open(id_file, "r") as f:
            builder_id = f.read()
        return builder_id

    def _getAndValidateConfig(self):
        config = nodepool_config.loadConfig(self._config_path)
        if self._secure_path:
            nodepool_config.loadSecureConfig(config, self._secure_path)
        if not config.zookeeper_servers.values():
            raise RuntimeError('No ZooKeeper servers specified in config.')
        if not config.imagesdir:
            raise RuntimeError('No images-dir specified in config.')
        return config

    # ======================================================================
    # Public methods
    # ======================================================================

    def start(self):
        '''
        Start the builder.

        The builder functionality is encapsulated within threads run
        by the NodePoolBuilder. This starts the needed sub-threads
        which will run forever until we tell them to stop.
        '''
        with self._start_lock:
            if self._running:
                raise exceptions.BuilderError('Cannot start, already running.')

            self._config = self._getAndValidateConfig()
            self._running = True

            builder_id_file = os.path.join(self._config.imagesdir,
                                           "builder_id.txt")
            builder_id = self._getBuilderID(builder_id_file)

            # All worker threads share a single ZooKeeper instance/connection.
            self.zk = zk.ZooKeeper(enable_cache=False)
            self.zk.connect(
                list(self._config.zookeeper_servers.values()),
                tls_cert=self._config.zookeeper_tls_cert,
                tls_key=self._config.zookeeper_tls_key,
                tls_ca=self._config.zookeeper_tls_ca)

            self.log.debug('Starting listener for build jobs')

            # Create build and upload worker objects
            for i in range(self._num_builders):
                w = BuildWorker(i, builder_id,
                                self._config_path, self._secure_path,
                                self.build_interval, self.zk)
                w.start()
                self._build_workers.append(w)

            for i in range(self._num_uploaders):
                w = UploadWorker(i, builder_id,
                                 self._config_path, self._secure_path,
                                 self.upload_interval, self.zk)
                w.start()
                self._upload_workers.append(w)

            if self.cleanup_interval > 0:
                self._janitor = CleanupWorker(
                    0, builder_id,
                    self._config_path, self._secure_path,
                    self.cleanup_interval, self.zk)
                self._janitor.start()

            # Wait until all threads are running. Otherwise, we have a race
            # on the worker _running attribute if shutdown() is called before
            # run() actually begins.
            workers = self._build_workers + self._upload_workers
            if self._janitor:
                workers += [self._janitor]
            while not all([
                x.running for x in (workers)]):
                time.sleep(0)

    def stop(self):
        '''
        Stop the builder.

        Signal the sub threads to begin the shutdown process. We don't
        want this method to return until the scheduler has successfully
        stopped all of its own threads.
        '''
        with self._start_lock:
            self.log.debug("Stopping. NodePoolBuilder shutting down workers")
            # Note we do not add the upload workers to this list intentionally.
            # The reason for this is that uploads can take many hours and there
            # is no good way to stop the blocking writes performed by the
            # uploads in order to join() below on a reasonable amount of time.
            # Killing the process will stop the upload then both the record
            # in zk and in the cloud will be deleted by any other running
            # builders or when this builder starts again.
            workers = self._build_workers
            if self._janitor:
                workers += [self._janitor]
            for worker in (workers):
                worker.shutdown()

        self._running = False

        self.log.debug('Waiting for jobs to complete')

        # Do not exit until all of our owned threads exit.
        for worker in (workers):
            worker.join()

        self.log.debug('Stopping providers')
        provider_manager.ProviderManager.stopProviders(self._config)

        self.log.debug('Terminating ZooKeeper connection')
        self.zk.disconnect()

        self.log.debug('Finished stopping')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools
import time

from nodepool import exceptions as npe
from nodepool import tests
from nodepool import zk
from nodepool.nodeutils import iterate_timeout


class TestZooKeeper(tests.DBTestCase):

    def setUp(self):
        super(TestZooKeeper, self).setUp()

    def test_buildZooKeeperHosts_single(self):
        hosts = [
            zk.ZooKeeperConnectionConfig('127.0.0.1', port=2181,
                                         chroot='/test1')
        ]
        self.assertEqual('127.0.0.1:2181/test1',
                         zk.buildZooKeeperHosts(hosts))

    def test_buildZooKeeperHosts_multiple(self):
        hosts = [
            zk.ZooKeeperConnectionConfig('127.0.0.1', port=2181,
                                         chroot='/test1'),
            zk.ZooKeeperConnectionConfig('127.0.0.2', port=2182,
                                         chroot='/test2')
        ]
        self.assertEqual('127.0.0.1:2181/test1,127.0.0.2:2182/test2',
                         zk.buildZooKeeperHosts(hosts))

    def test_buildZooKeeperHosts_ipv6(self):
        hosts = [
            zk.ZooKeeperConnectionConfig(
                '2001:4800:7817:103:be76:4eff:fe04:e359', port=2181,
                chroot='/test1'),
            zk.ZooKeeperConnectionConfig(
                '[2002:4800:7817:103:be76:4eff:fe04:e359]', port=2181,
                chroot='/test2'),
            zk.ZooKeeperConnectionConfig('127.0.0.2', port=2182,
                                         chroot='/test3')
        ]
        self.assertEqual((
            '[2001:4800:7817:103:be76:4eff:fe04:e359]:2181/test1,'
            '[2002:4800:7817:103:be76:4eff:fe04:e359]:2181/test2,'
            '127.0.0.2:2182/test3'
        ), zk.buildZooKeeperHosts(hosts))

    def test_imageBuildLock(self):
        path = self.zk._imageBuildLockPath("ubuntu-trusty")
        with self.zk.imageBuildLock("ubuntu-trusty", blocking=False):
            self.assertIsNotNone(self.zk.client.exists(path))

    def test_imageBuildLock_exception_nonblocking(self):
        image = "ubuntu-trusty"
        with self.zk.imageBuildLock(image, blocking=False):
            with testtools.ExpectedException(
                npe.ZKLockException, "Did not get lock on .*"
            ):
                with self.zk.imageBuildLock(image, blocking=False):
                    pass

    def test_imageBuildLock_exception_blocking(self):
        image = "ubuntu-trusty"
        with self.zk.imageBuildLock(image, blocking=False):
            with testtools.ExpectedException(npe.TimeoutException):
                with self.zk.imageBuildLock(image, blocking=True, timeout=1):
                    pass

    def test_imageBuildNumberLock(self):
        path = self.zk._imageBuildNumberLockPath("ubuntu-trusty", "0000")
        with self.zk.imageBuildNumberLock(
            "ubuntu-trusty", "0000", blocking=False
        ):
            self.assertIsNotNone(self.zk.client.exists(path))

    def test_imageBuildNumberLock_exception_nonblocking(self):
        image = "ubuntu-trusty"
        bnum = "0000000000"
        with self.zk.imageBuildNumberLock(image, bnum, blocking=False):
            with testtools.ExpectedException(
                npe.ZKLockException, "Did not get lock on .*"
            ):
                with self.zk.imageBuildNumberLock(image, bnum, blocking=False):
                    pass

    def test_imageBuildNumberLock_exception_blocking(self):
        image = "ubuntu-trusty"
        bnum = "0000000000"
        with self.zk.imageBuildNumberLock(image, bnum, blocking=False):
            with testtools.ExpectedException(npe.TimeoutException):
                with self.zk.imageBuildNumberLock(
                    image, bnum, blocking=True, timeout=1
                ):
                    pass

    def test_imageUploadNumberLock_delete(self):
        '''
        Test that deleting an image upload number while we hold the lock
        on it works properly.
        '''
        upload = zk.ImageUpload()
        upload.image_name = "ubuntu-trusty"
        upload.build_id = "0000000003"
        upload.provider_name = "providerA"
        upload.id = "0000000001"
        path = self.zk._imageUploadNumberLockPath(upload.image_name,
                                                  upload.build_id,
                                                  upload.provider_name,
                                                  upload.id)
        with self.zk.imageUploadNumberLock(upload, blocking=False):
            self.assertIsNotNone(self.zk.client.exists(path))
            self.zk.deleteUpload(upload.image_name,
                                 upload.build_id,
                                 upload.provider_name,
                                 upload.id)
        self.assertIsNone(self.zk.client.exists(path))

    def test_imageUploadNumberLock(self):
        upload = zk.ImageUpload()
        upload.image_name = "ubuntu-trusty"
        upload.build_id = "0000000003"
        upload.provider_name = "providerA"
        upload.id = "0000000001"
        path = self.zk._imageUploadNumberLockPath(upload.image_name,
                                                  upload.build_id,
                                                  upload.provider_name,
                                                  upload.id)
        with self.zk.imageUploadNumberLock(upload, blocking=False):
            self.assertIsNotNone(self.zk.client.exists(path))

    def test_imageUploadNumberLock_exception_nonblocking(self):
        upload = zk.ImageUpload()
        upload.image_name = "ubuntu-trusty"
        upload.build_id = "0000000003"
        upload.provider_name = "providerA"
        upload.id = "0000000001"
        with self.zk.imageUploadNumberLock(upload, blocking=False):
            with testtools.ExpectedException(
                npe.ZKLockException, "Did not get lock on .*"
            ):
                with self.zk.imageUploadNumberLock(upload, blocking=False):
                    pass

    def test_imageUploadNumberLock_exception_blocking(self):
        upload = zk.ImageUpload()
        upload.image_name = "ubuntu-trusty"
        upload.build_id = "0000000003"
        upload.provider_name = "providerA"
        upload.id = "0000000001"
        with self.zk.imageUploadNumberLock(upload, blocking=False):
            with testtools.ExpectedException(npe.TimeoutException):
                with self.zk.imageUploadNumberLock(
                    upload, blocking=True, timeout=1
                ):
                    pass

    def test_imageUploadLock(self):
        path = self.zk._imageUploadLockPath("ubuntu-trusty", "0000", "prov1")
        with self.zk.imageUploadLock("ubuntu-trusty", "0000", "prov1",
                                     blocking=False):
            self.assertIsNotNone(self.zk.client.exists(path))

    def test_imageUploadLock_exception_nonblocking(self):
        image = "ubuntu-trusty"
        bnum = "0000000000"
        prov = "rax"
        with self.zk.imageUploadLock(image, bnum, prov, blocking=False):
            with testtools.ExpectedException(
                npe.ZKLockException, "Did not get lock on .*"
            ):
                with self.zk.imageUploadLock(image, bnum, prov,
                                             blocking=False):
                    pass

    def test_imageUploadLock_exception_blocking(self):
        image = "ubuntu-trusty"
        bnum = "0000000000"
        prov = "rax"
        with self.zk.imageUploadLock(image, bnum, prov, blocking=False):
            with testtools.ExpectedException(npe.TimeoutException):
                with self.zk.imageUploadLock(image, bnum, prov,
                                             blocking=True, timeout=1):
                    pass

    def test_storeBuild(self):
        image = "ubuntu-trusty"
        b1 = self.zk.storeBuild(image, zk.ImageBuild())
        b2 = self.zk.storeBuild(image, zk.ImageBuild())
        self.assertLess(int(b1), int(b2))

    def test_store_and_get_build(self):
        image = "ubuntu-trusty"
        orig_data = zk.ImageBuild()
        orig_data.builder = 'host.domain.com'
        orig_data.builder_id = 'ABC-123'
        orig_data.state = zk.READY
        with self.zk.imageBuildLock(image, blocking=True, timeout=1):
            build_num = self.zk.storeBuild(image, orig_data)

        data = self.zk.getBuild(image, build_num)
        self.assertEqual(orig_data.builder, data.builder)
        self.assertEqual(orig_data.builder_id, data.builder_id)
        self.assertEqual(orig_data.state, data.state)
        self.assertEqual(orig_data.state_time, data.state_time)
        self.assertEqual(build_num, data.id)
        self.assertEqual(self.zk.getImageNames(), ["ubuntu-trusty"])
        self.assertEqual(self.zk.getBuildNumbers("ubuntu-trusty"), [build_num])

    def test_getImageNames_not_found(self):
        self.assertEqual(self.zk.getImageNames(), [])

    def test_getBuildNumbers_not_found(self):
        self.assertEqual(self.zk.getBuildNumbers("ubuntu-trusty"), [])

    def test_getBuildProviders_not_found(self):
        self.assertEqual(self.zk.getBuildProviders(
            "ubuntu-trusty", "0000000000"), [])

    def test_getImageUploadNumbers_not_found(self):
        self.assertEqual(self.zk.getImageUploadNumbers(
            "ubuntu-trusty", "0000000000", "rax"), [])

    def test_getBuild_not_found(self):
        self.assertIsNone(self.zk.getBuild("ubuntu-trusty", "0000000000"))

    def test_getImageUpload_not_found(self):
        self.assertIsNone(
            self.zk.getImageUpload("trusty", "0001", "rax", "0000000001")
        )

    def test_storeImageUpload(self):
        image = "ubuntu-trusty"
        provider = "rax"
        bnum = self.zk.storeBuild(image, zk.ImageBuild())
        up1 = self.zk.storeImageUpload(image, bnum, provider, zk.ImageUpload())
        up2 = self.zk.storeImageUpload(image, bnum, provider, zk.ImageUpload())
        self.assertLess(int(up1), int(up2))

    def test_storeImageUpload_invalid_build(self):
        image = "ubuntu-trusty"
        build_number = "0000000001"
        provider = "rax"
        orig_data = zk.ImageUpload()

        with testtools.ExpectedException(
            npe.ZKException, "Cannot find build .*"
        ):
            self.zk.storeImageUpload(image, build_number, provider, orig_data)

    def test_store_and_get_image_upload(self):
        image = "ubuntu-trusty"
        provider = "rax"
        orig_data = zk.ImageUpload()
        orig_data.external_id = "deadbeef"
        orig_data.state = zk.READY
        orig_data.format = "qcow2"

        build_number = self.zk.storeBuild(image, zk.ImageBuild())
        upload_id = self.zk.storeImageUpload(image, build_number, provider,
                                             orig_data)
        data = self.zk.getImageUpload(image, build_number, provider, upload_id)

        self.assertEqual(upload_id, data.id)
        self.assertEqual(orig_data.external_id, data.external_id)
        self.assertEqual(orig_data.state, data.state)
        self.assertEqual(orig_data.state_time, data.state_time)
        self.assertEqual(orig_data.format, data.format)
        self.assertEqual(self.zk.getBuildProviders("ubuntu-trusty",
                                                   build_number),
                         [provider])
        self.assertEqual(self.zk.getImageUploadNumbers("ubuntu-trusty",
                                                       build_number,
                                                       provider),
                         [upload_id])

    def test_build_request(self):
        '''Test the build request API methods (has/submit/remove)'''
        image = "ubuntu-trusty"
        self.zk.submitBuildRequest(image)
        self.assertTrue(self.zk.hasBuildRequest(image))
        self.zk.removeBuildRequest(image)
        self.assertFalse(self.zk.hasBuildRequest(image))

    def test_getMostRecentBuilds(self):
        image = "ubuntu-trusty"
        v1 = {'state': zk.READY, 'state_time': int(time.time())}
        v2 = {'state': zk.READY, 'state_time': v1['state_time'] + 10}
        v3 = {'state': zk.READY, 'state_time': v1['state_time'] + 20}
        v4 = {'state': zk.DELETING, 'state_time': v2['state_time'] + 10}
        self.zk.storeBuild(image, zk.ImageBuild.fromDict(v1))
        v2_id = self.zk.storeBuild(image, zk.ImageBuild.fromDict(v2))
        v3_id = self.zk.storeBuild(image, zk.ImageBuild.fromDict(v3))
        self.zk.storeBuild(image, zk.ImageBuild.fromDict(v4))

        # v2 and v3 should be the 2 most recent 'ready' builds
        matches = self.zk.getMostRecentBuilds(2, image, zk.READY)
        self.assertEqual(2, len(matches))

        # Should be in descending order, according to state_time
        self.assertEqual(matches[0].id, v3_id)
        self.assertEqual(matches[1].id, v2_id)

    def test_getMostRecentBuildImageUploads_with_state(self):
        image = "ubuntu-trusty"
        provider = "rax"
        build = {'state': zk.READY, 'state_time': int(time.time())}
        up1 = zk.ImageUpload()
        up1.state = zk.READY
        up2 = zk.ImageUpload()
        up2.state = zk.READY
        up2.state_time = up1.state_time + 10
        up3 = zk.ImageUpload()
        up3.state = zk.DELETING
        up3.state_time = up2.state_time + 10

        bnum = self.zk.storeBuild(image, zk.ImageBuild.fromDict(build))
        self.zk.storeImageUpload(image, bnum, provider, up1)
        up2_id = self.zk.storeImageUpload(image, bnum, provider, up2)
        self.zk.storeImageUpload(image, bnum, provider, up3)

        # up2 should be the most recent 'ready' upload
        data = self.zk.getMostRecentBuildImageUploads(
            1, image, bnum, provider, zk.READY)
        self.assertNotEqual([], data)
        self.assertEqual(1, len(data))
        self.assertEqual(data[0].id, up2_id)

    def test_getMostRecentBuildImageUploads_any_state(self):
        image = "ubuntu-trusty"
        provider = "rax"
        build = {'state': zk.READY, 'state_time': int(time.time())}
        up1 = zk.ImageUpload()
        up1.state = zk.READY
        up2 = zk.ImageUpload()
        up2.state = zk.READY
        up2.state_time = up1.state_time + 10
        up3 = zk.ImageUpload()
        up3.state = zk.UPLOADING
        up3.state_time = up2.state_time + 10

        bnum = self.zk.storeBuild(image, zk.ImageBuild.fromDict(build))
        self.zk.storeImageUpload(image, bnum, provider, up1)
        self.zk.storeImageUpload(image, bnum, provider, up2)
        up3_id = self.zk.storeImageUpload(image, bnum, provider, up3)

        # up3 should be the most recent upload, regardless of state
        data = self.zk.getMostRecentBuildImageUploads(
            1, image, bnum, provider, None)
        self.assertNotEqual([], data)
        self.assertEqual(1, len(data))
        self.assertEqual(data[0].id, up3_id)

    def test_getMostRecentImageUpload(self):
        image = "ubuntu-trusty"
        provider = "rax"

        build1 = zk.ImageBuild()
        build1.state = zk.READY
        build2 = zk.ImageBuild()
        build2.state = zk.READY
        build2.state_time = build1.state_time + 10

        bnum1 = self.zk.storeBuild(image, build1)
        bnum2 = self.zk.storeBuild(image, build2)

        upload1 = zk.ImageUpload()
        upload1.state = zk.READY
        upload2 = zk.ImageUpload()
        upload2.state = zk.READY
        upload2.state_time = upload1.state_time + 10

        self.zk.storeImageUpload(image, bnum1, provider, upload1)
        self.zk.storeImageUpload(image, bnum2, provider, upload2)

        d = self.zk.getMostRecentImageUpload(image, provider, zk.READY)
        self.assertEqual(upload2.state_time, d.state_time)

    def test_getProviderUploads(self):
        image = "ubuntu-trusty"
        provider = "rax"

        build = zk.ImageBuild()
        build.state = zk.READY
        bnum = self.zk.storeBuild(image, build)

        upload = zk.ImageUpload()
        upload.state = zk.READY
        upnum = self.zk.storeImageUpload(image, bnum, provider, upload)

        d = self.zk.getProviderUploads(provider)
        self.assertIn(image, d)
        self.assertIn(bnum, d[image])
        self.assertEqual(1, len(d[image][bnum]))
        self.assertIsInstance(d[image][bnum][0], zk.ImageUpload)
        self.assertEqual(upnum, d[image][bnum][0].id)

    def test_getBuilds_any(self):
        image = "ubuntu-trusty"
        path = self.zk._imageBuildsPath(image)
        v1 = zk.ImageBuild()
        v1.state = zk.READY
        v2 = zk.ImageBuild()
        v2.state = zk.BUILDING
        v3 = zk.ImageBuild()
        v3.state = zk.FAILED
        v4 = zk.ImageBuild()
        v4.state = zk.DELETING
        self.zk.client.create(path + "/1", value=v1.serialize(), makepath=True)
        self.zk.client.create(path + "/2", value=v2.serialize(), makepath=True)
        self.zk.client.create(path + "/3", value=v3.serialize(), makepath=True)
        self.zk.client.create(path + "/4", value=v4.serialize(), makepath=True)
        self.zk.client.create(path + "/lock", makepath=True)

        matches = self.zk.getBuilds(image, None)
        self.assertEqual(4, len(matches))

    def test_getBuilds(self):
        image = "ubuntu-trusty"
        path = self.zk._imageBuildsPath(image)
        v1 = zk.ImageBuild()
        v1.state = zk.READY
        v2 = zk.ImageBuild()
        v2.state = zk.BUILDING
        v3 = zk.ImageBuild()
        v3.state = zk.FAILED
        v4 = zk.ImageBuild()
        v4.state = zk.DELETING
        self.zk.client.create(path + "/1", value=v1.serialize(), makepath=True)
        self.zk.client.create(path + "/2", value=v2.serialize(), makepath=True)
        self.zk.client.create(path + "/3", value=v3.serialize(), makepath=True)
        self.zk.client.create(path + "/4", value=v4.serialize(), makepath=True)
        self.zk.client.create(path + "/lock", makepath=True)

        matches = self.zk.getBuilds(image, [zk.DELETING, zk.FAILED])
        self.assertEqual(2, len(matches))

    def test_getUploads(self):
        path = self.zk._imageUploadPath("trusty", "000", "rax")
        v1 = zk.ImageUpload()
        v1.state = zk.READY
        v2 = zk.ImageUpload()
        v2.state = zk.UPLOADING
        v3 = zk.ImageUpload()
        v3.state = zk.FAILED
        v4 = zk.ImageUpload()
        v4.state = zk.DELETING
        self.zk.client.create(path + "/1", value=v1.serialize(), makepath=True)
        self.zk.client.create(path + "/2", value=v2.serialize(), makepath=True)
        self.zk.client.create(path + "/3", value=v3.serialize(), makepath=True)
        self.zk.client.create(path + "/4", value=v4.serialize(), makepath=True)
        self.zk.client.create(path + "/lock", makepath=True)

        matches = self.zk.getUploads("trusty", "000", "rax",
                                     [zk.DELETING, zk.FAILED])
        self.assertEqual(2, len(matches))

    def test_getUploads_any(self):
        path = self.zk._imageUploadPath("trusty", "000", "rax")
        v1 = zk.ImageUpload()
        v1.state = zk.READY
        v2 = zk.ImageUpload()
        v2.state = zk.UPLOADING
        v3 = zk.ImageUpload()
        v3.state = zk.FAILED
        v4 = zk.ImageUpload()
        v4.state = zk.DELETING
        self.zk.client.create(path + "/1", value=v1.serialize(), makepath=True)
        self.zk.client.create(path + "/2", value=v2.serialize(), makepath=True)
        self.zk.client.create(path + "/3", value=v3.serialize(), makepath=True)
        self.zk.client.create(path + "/4", value=v4.serialize(), makepath=True)
        self.zk.client.create(path + "/lock", makepath=True)

        matches = self.zk.getUploads("trusty", "000", "rax", None)
        self.assertEqual(4, len(matches))

    def test_deleteBuild(self):
        image = 'trusty'
        build = zk.ImageBuild()
        build.state = zk.READY
        bnum = self.zk.storeBuild(image, build)
        self.assertTrue(self.zk.deleteBuild(image, bnum))

    def test_deleteBuild_with_uploads(self):
        image = 'trusty'
        provider = 'rax'

        build = zk.ImageBuild()
        build.state = zk.READY
        bnum = self.zk.storeBuild(image, build)

        upload = zk.ImageUpload()
        upload.state = zk.READY
        self.zk.storeImageUpload(image, bnum, provider, upload)

        self.assertFalse(self.zk.deleteBuild(image, bnum))

    def test_deleteUpload(self):
        path = self.zk._imageUploadPath("trusty", "000", "rax") + "/000001"
        self.zk.client.create(path, makepath=True)
        self.zk.deleteUpload("trusty", "000", "rax", "000001")
        self.assertIsNone(self.zk.client.exists(path))

    def test_registerLauncher(self):
        launcher = zk.Launcher()
        launcher.id = "launcher-000-001"
        self.zk.registerLauncher(launcher)
        launchers = self.zk.getRegisteredLaunchers()
        self.assertEqual(1, len(launchers))
        self.assertEqual(launcher.id, launchers[0].id)

    def test_registerLauncher_safe_repeat(self):
        launcher = zk.Launcher()
        launcher.id = "launcher-000-001"
        self.zk.registerLauncher(launcher)
        self.zk.registerLauncher(launcher)
        launchers = self.zk.getRegisteredLaunchers()
        self.assertEqual(1, len(launchers))
        self.assertEqual(launcher.id, launchers[0].id)

    def test_getNodeRequests_empty(self):
        self.assertEqual([], self.zk.getNodeRequests())

    def test_getNodeRequests(self):
        r1 = self.zk._requestPath("500-123")
        r2 = self.zk._requestPath("100-456")
        r3 = self.zk._requestPath("100-123")
        r4 = self.zk._requestPath("400-123")
        self.zk.client.create(r1, makepath=True, ephemeral=True)
        self.zk.client.create(r2, makepath=True, ephemeral=True)
        self.zk.client.create(r3, makepath=True, ephemeral=True)
        self.zk.client.create(r4, makepath=True, ephemeral=True)

        self.assertEqual(
            ["100-123", "100-456", "400-123", "500-123"],
            self.zk.getNodeRequests()
        )

    def test_getNodeRequest(self):
        r = zk.NodeRequest("500-123")
        r.state = zk.REQUESTED
        path = self.zk._requestPath(r.id)
        self.zk.client.create(path, value=r.serialize(),
                              makepath=True, ephemeral=True)
        o = self.zk.getNodeRequest(r.id)
        self.assertIsInstance(o, zk.NodeRequest)
        self.assertEqual(r.id, o.id)

    def test_getNodeRequest_not_found(self):
        self.assertIsNone(self.zk.getNodeRequest("invalid"))

    def test_getNodes(self):
        self.zk.client.create(self.zk._nodePath('100'), makepath=True)
        self.zk.client.create(self.zk._nodePath('200'), makepath=True)
        nodes = self.zk.getNodes()
        self.assertIn('100', nodes)
        self.assertIn('200', nodes)

    def test_getNode(self):
        n = zk.Node('100')
        n.state = zk.BUILDING
        path = self.zk._nodePath(n.id)
        self.zk.client.create(path, value=n.serialize(), makepath=True)
        o = self.zk.getNode(n.id)
        self.assertIsInstance(o, zk.Node)
        self.assertEqual(n.id, o.id)

    def test_getNode_not_found(self):
        self.assertIsNone(self.zk.getNode("invalid"))

    def test_lockNode_multi(self):
        node = zk.Node('100')
        self.zk.lockNode(node)
        with testtools.ExpectedException(
            npe.ZKLockException, "Did not get lock on .*"
        ):
            self.zk.lockNode(node, blocking=False)

    def test_lockNode_unlockNode(self):
        node = zk.Node('100')
        self.zk.lockNode(node)
        self.assertIsNotNone(node.lock)
        self.assertIsNotNone(
            self.zk.client.exists(self.zk._nodeLockPath(node.id))
        )
        self.zk.unlockNode(node)
        self.assertIsNone(node.lock)

    def test_unlockNode_not_locked(self):
        node = zk.Node('100')
        with testtools.ExpectedException(npe.ZKLockException):
            self.zk.unlockNode(node)

    def _create_node(self):
        node = zk.Node()
        node.state = zk.BUILDING
        node.provider = 'rax'

        self.assertIsNone(node.id)
        self.zk.storeNode(node)
        self.assertIsNotNone(node.id)
        self.assertIsNotNone(
            self.zk.client.exists(self.zk._nodePath(node.id))
        )
        return node

    def test_storeNode(self):
        node = self._create_node()
        node2 = self.zk.getNode(node.id)
        self.assertEqual(node, node2)

    def test_storeNode_update(self):
        node = self._create_node()
        node.state = zk.READY
        self.zk.storeNode(node)
        node2 = self.zk.getNode(node.id)
        self.assertEqual(node, node2)

    def _create_node_request(self):
        req = zk.NodeRequest()
        req.state = zk.REQUESTED
        req.node_types.append('label1')
        self.zk.storeNodeRequest(req)
        self.assertIsNotNone(
            self.zk.client.exists(self.zk._requestPath(req.id))
        )
        return req

    def test_storeNodeRequest(self):
        req = self._create_node_request()
        self.assertNotEqual(req.event_id, None)
        req2 = self.zk.getNodeRequest(req.id)
        self.assertEqual(req, req2)

    def test_storeNodeRequest_update(self):
        req = self._create_node_request()
        req.state = zk.FULFILLED
        self.zk.storeNodeRequest(req)
        self.assertIsNotNone(req.id)
        req2 = self.zk.getNodeRequest(req.id)
        self.assertEqual(req, req2)

    def test_deleteNodeRequest(self):
        req = self._create_node_request()
        self.zk.deleteNodeRequest(req)
        self.assertIsNone(
            self.zk.client.exists(self.zk._requestPath(req.id))
        )

    def test_deleteNode(self):
        n1 = self._create_node()
        self.zk.deleteNode(n1)
        self.assertIsNone(
            self.zk.client.exists(self.zk._nodePath(n1.id))
        )

    def test_getReadyNodesOfTypes(self):
        n1 = self._create_node()
        n1.type = 'label1'
        n1.state = zk.READY
        self.zk.storeNode(n1)
        n2 = self._create_node()
        n2.state = zk.READY
        n2.type = 'label1'
        self.zk.storeNode(n2)
        n3 = self._create_node()
        n3.state = zk.READY
        n3.type = 'label2'
        self.zk.storeNode(n3)

        r = self.zk.getReadyNodesOfTypes(['label1'], cached=False)
        self.assertIn('label1', r)
        self.assertEqual(2, len(r['label1']))
        self.assertIn(n1, r['label1'])
        self.assertIn(n2, r['label1'])

    def test_getReadyNodesOfTypes_multilabel(self):
        n1 = self._create_node()
        n1.type = 'label1'
        n1.state = zk.READY
        self.zk.storeNode(n1)
        n2 = self._create_node()
        n2.state = zk.READY
        n2.type = ['label1', 'label3']
        self.zk.storeNode(n2)
        n3 = self._create_node()
        n3.state = zk.READY
        n3.type = 'label2'
        self.zk.storeNode(n3)

        r = self.zk.getReadyNodesOfTypes(['label1', 'label3'], cached=False)
        self.assertIn('label1', r)
        self.assertIn('label3', r)
        self.assertEqual(2, len(r['label1']))
        self.assertEqual(1, len(r['label3']))
        self.assertIn(n1, r['label1'])
        self.assertIn(n2, r['label1'])
        self.assertIn(n2, r['label3'])

    def test_nodeIterator(self):
        n1 = self._create_node()
        i = self.zk.nodeIterator(cached=False)
        self.assertEqual(n1, next(i))
        with testtools.ExpectedException(StopIteration):
            next(i)

    def test_getNodeRequestLockIDs(self):
        req = self._create_node_request()
        self.zk.lockNodeRequest(req, blocking=False)
        lock_ids = self.zk.getNodeRequestLockIDs()
        self.assertEqual(1, len(lock_ids))
        self.assertEqual(req.id, lock_ids[0])
        self.zk.unlockNodeRequest(req)
        self.zk.deleteNodeRequest(req)

    def test_getNodeRequestLockStats(self):
        req = self._create_node_request()
        self.zk.lockNodeRequest(req, blocking=False)
        lock_stats = self.zk.getNodeRequestLockStats(req.id)
        self.assertEqual(lock_stats.lock_id, req.id)
        self.assertIsNotNone(lock_stats.stat)
        self.zk.unlockNodeRequest(req)
        self.zk.deleteNodeRequest(req)

    def test_nodeRequestLockStatsIterator(self):
        req = self._create_node_request()
        self.zk.lockNodeRequest(req, blocking=False)
        i = self.zk.nodeRequestLockStatsIterator()
        self.assertEqual(zk.NodeRequestLockStats(req.id), next(i))
        with testtools.ExpectedException(StopIteration):
            next(i)
        self.zk.unlockNodeRequest(req)
        self.zk.deleteNodeRequest(req)

    def test_nodeRequestIterator(self):
        req = self._create_node_request()
        self.zk.lockNodeRequest(req, blocking=False)
        i = self.zk.nodeRequestIterator()
        self.assertEqual(req, next(i))
        with testtools.ExpectedException(StopIteration):
            next(i)
        self.zk.unlockNodeRequest(req)
        self.zk.deleteNodeRequest(req)

    def test_deleteNodeRequestLock(self):
        req = self._create_node_request()
        self.zk.lockNodeRequest(req, blocking=False)
        self.zk.unlockNodeRequest(req)
        self.zk.deleteNodeRequest(req)

        # We expect the lock to linger even after the request is deleted
        lock_ids = self.zk.getNodeRequestLockIDs()
        self.assertEqual(1, len(lock_ids))
        self.assertEqual(req.id, lock_ids[0])
        self.zk.deleteNodeRequestLock(lock_ids[0])
        self.assertEqual([], self.zk.getNodeRequestLockIDs())

    def test_node_caching(self):
        '''
        Test that node iteration using both cached and uncached calls
        produces identical results.
        '''
        # Test new node in node set
        n1 = self._create_node()

        # uncached
        a1 = self.zk.nodeIterator(cached=False)
        self.assertEqual(n1, next(a1))

        # cached
        a2 = self.zk.nodeIterator(cached=True)
        self.assertEqual(n1, next(a2))
        with testtools.ExpectedException(StopIteration):
            next(a2)

        # Test modification of existing node set
        n1.state = zk.HOLD
        n1.label = "oompaloompa"
        self.zk.storeNode(n1)

        # uncached
        b1 = self.zk.nodeIterator(cached=False)
        self.assertEqual(n1, next(b1))

        # cached
        for _ in iterate_timeout(10, Exception,
                                 "cached node equals original node"):
            b2 = self.zk.nodeIterator(cached=True)
            if n1 == next(b2):
                break


class TestZKModel(tests.BaseTestCase):

    def setUp(self):
        super(TestZKModel, self).setUp()

    def test_BaseModel_bad_id(self):
        with testtools.ExpectedException(
            TypeError, "'id' attribute must be a string type"
        ):
            zk.BaseModel(123)

    def test_BaseModel_bad_state(self):
        with testtools.ExpectedException(
            TypeError, "'blah' is not a valid state"
        ):
            o = zk.BaseModel('0001')
            o.state = 'blah'

    def test_BaseModel_toDict(self):
        o = zk.BaseModel('0001')
        d = o.toDict()
        self.assertNotIn('id', d)

    def test_ImageBuild_toDict(self):
        o = zk.ImageBuild('0001')
        o.state = zk.BUILDING
        o.builder = 'localhost'
        o.builder_id = 'ABC-123'
        o.formats = ['qemu', 'raw']

        d = o.toDict()
        self.assertNotIn('id', d)
        self.assertEqual(o.state, d['state'])
        self.assertIsNotNone(d['state_time'])
        self.assertEqual(','.join(o.formats), d['formats'])
        self.assertEqual(o.builder, d['builder'])
        self.assertEqual(o.builder_id, d['builder_id'])

    def test_ImageBuild_fromDict(self):
        now = int(time.time())
        d_id = '0001'
        d = {
            'builder': 'localhost',
            'builder_id': 'ABC-123',
            'formats': 'qemu,raw',
            'state': zk.BUILDING,
            'state_time': now
        }

        o = zk.ImageBuild.fromDict(d, d_id)
        self.assertEqual(o.id, d_id)
        self.assertEqual(o.state, d['state'])
        self.assertEqual(o.state_time, d['state_time'])
        self.assertEqual(o.builder, d['builder'])
        self.assertEqual(o.builder_id, d['builder_id'])
        self.assertEqual(o.formats, d['formats'].split(','))

    def test_ImageUpload_toDict(self):
        o = zk.ImageUpload('0001', '0003')
        o.state = zk.UPLOADING
        o.external_id = 'DEADBEEF'
        o.external_name = 'trusty'
        o.format = 'qcow2'

        d = o.toDict()
        self.assertNotIn('id', d)
        self.assertNotIn('build_id', d)
        self.assertNotIn('provider_name', d)
        self.assertNotIn('image_name', d)
        self.assertEqual(o.state, d['state'])
        self.assertEqual(o.state_time, d['state_time'])
        self.assertEqual(o.external_id, d['external_id'])
        self.assertEqual(o.external_name, d['external_name'])
        self.assertEqual(o.format, d['format'])

    def test_ImageUpload_fromDict(self):
        now = int(time.time())
        upload_id = '0001'
        build_id = '0003'
        d = {
            'external_id': 'DEADBEEF',
            'external_name': 'trusty',
            'format': 'qcow2',
            'state': zk.READY,
            'state_time': now
        }

        o = zk.ImageUpload.fromDict(d, build_id, 'rax', 'trusty', upload_id)
        self.assertEqual(o.id, upload_id)
        self.assertEqual(o.build_id, build_id)
        self.assertEqual(o.provider_name, 'rax')
        self.assertEqual(o.image_name, 'trusty')
        self.assertEqual(o.state, d['state'])
        self.assertEqual(o.state_time, d['state_time'])
        self.assertEqual(o.external_id, d['external_id'])
        self.assertEqual(o.external_name, d['external_name'])
        self.assertEqual(o.format, d['format'])

    def test_NodeRequest_toDict(self):
        o = zk.NodeRequest("500-123")
        o.declined_by.append("abc")
        o.node_types.append('trusty')
        o.nodes.append('100')
        o.reuse = False
        o.requestor = 'zuul'
        o.provider = "test"
        o.relative_priority = 1
        o.event_id = "deadbeef"
        d = o.toDict()
        self.assertNotIn('id', d)
        self.assertIn('state', d)
        self.assertIn('state_time', d)
        self.assertEqual(d['declined_by'], o.declined_by)
        self.assertEqual(d['node_types'], o.node_types)
        self.assertEqual(d['nodes'], o.nodes)
        self.assertEqual(d['reuse'], o.reuse)
        self.assertEqual(d['requestor'], o.requestor)
        self.assertEqual(d['provider'], o.provider)
        self.assertEqual(d['relative_priority'], o.relative_priority)
        self.assertEqual(d['event_id'], o.event_id)

    def test_NodeRequest_fromDict(self):
        now = int(time.time())
        req_id = "500-123"
        d = {
            'state': zk.REQUESTED,
            'state_time': now,
            'declined_by': ['abc'],
            'node_types': ['trusty'],
            'nodes': ['100'],
            'reuse': False,
            'requestor': 'zuul',
            'provider': 'test',
            'relative_priority': 1,
            'event_id': 'deadbeef',
        }

        o = zk.NodeRequest.fromDict(d, req_id)
        self.assertEqual(o.id, req_id)
        self.assertEqual(o.state, d['state'])
        self.assertEqual(o.state_time, d['state_time'])
        self.assertEqual(o.declined_by, d['declined_by'])
        self.assertEqual(o.node_types, d['node_types'])
        self.assertEqual(o.nodes, d['nodes'])
        self.assertEqual(o.reuse, d['reuse'])
        self.assertEqual(o.requestor, d['requestor'])
        self.assertEqual(o.provider, d['provider'])
        self.assertEqual(o.relative_priority, d['relative_priority'])
        self.assertEqual(o.event_id, d['event_id'])

    def test_Node_toDict(self):
        o = zk.Node('123')
        o.state = zk.INIT
        o.provider = 'rax'
        o.type = 'trusty'
        o.allocated_to = '456-789'
        o.az = 'RegionOne'
        o.region = 'fake-region'
        o.public_ipv4 = '<ipv4>'
        o.private_ipv4 = '<pvt-ipv4>'
        o.public_ipv6 = '<ipv6>'
        o.host_id = 'fake-host-id'
        o.image_id = 'image-id'
        o.launcher = 'launcher-id'
        o.external_id = 'ABCD'
        o.hostname = 'xyz'
        o.comment = 'comment'
        o.hold_job = 'hold job'
        o.host_keys = ['key1', 'key2']
        o.attributes = {'executor-zone': 'vpn'}

        d = o.toDict()
        self.assertNotIn('id', d)
        self.assertEqual(d['state'], o.state)
        self.assertIn('state_time', d)
        self.assertIn('created_time', d)
        self.assertEqual(d['provider'], o.provider)
        self.assertEqual(d['type'], o.type)
        self.assertEqual(d['allocated_to'], o.allocated_to)
        self.assertEqual(d['az'], o.az)
        self.assertEqual(d['region'], o.region)
        self.assertEqual(d['public_ipv4'], o.public_ipv4)
        self.assertEqual(d['private_ipv4'], o.private_ipv4)
        self.assertEqual(d['public_ipv6'], o.public_ipv6)
        self.assertEqual(d['host_id'], o.host_id)
        self.assertEqual(d['image_id'], o.image_id)
        self.assertEqual(d['launcher'], o.launcher)
        self.assertEqual(d['external_id'], o.external_id)
        self.assertEqual(d['hostname'], o.hostname)
        self.assertEqual(d['comment'], o.comment)
        self.assertEqual(d['hold_job'], o.hold_job)
        self.assertEqual(d['host_keys'], o.host_keys)
        self.assertEqual(d['attributes'], o.attributes)

    def test_Node_fromDict(self):
        now = int(time.time())
        node_id = '123'
        d = {
            'state': zk.READY,
            'state_time': now,
            'created_time': now - 2,
            'provider': 'rax',
            'type': ['trusty'],
            'allocated_to': '456-789',
            'az': 'RegionOne',
            'region': 'fake-region',
            'public_ipv4': '<ipv4>',
            'private_ipv4': '<pvt-ipv4>',
            'public_ipv6': '<ipv6>',
            'host_id': 'fake-host-id',
            'image_id': 'image-id',
            'launcher': 'launcher-id',
            'external_id': 'ABCD',
            'hostname': 'xyz',
            'comment': 'comment',
            'hold_job': 'hold job',
            'host_keys': ['key1', 'key2'],
            'connection_port': 22022,
            'attributes': {'executor-zone': 'vpn'},
        }

        o = zk.Node.fromDict(d, node_id)
        self.assertEqual(o.id, node_id)
        self.assertEqual(o.state, d['state'])
        self.assertEqual(o.state_time, d['state_time'])
        self.assertEqual(o.created_time, d['created_time'])
        self.assertEqual(o.provider, d['provider'])
        self.assertEqual(o.type, d['type'])
        self.assertEqual(o.allocated_to, d['allocated_to'])
        self.assertEqual(o.az, d['az'])
        self.assertEqual(o.region, d['region'])
        self.assertEqual(o.public_ipv4, d['public_ipv4'])
        self.assertEqual(o.private_ipv4, d['private_ipv4'])
        self.assertEqual(o.public_ipv6, d['public_ipv6'])
        self.assertEqual(o.host_id, d['host_id'])
        self.assertEqual(o.image_id, d['image_id'])
        self.assertEqual(o.launcher, d['launcher'])
        self.assertEqual(o.external_id, d['external_id'])
        self.assertEqual(o.hostname, d['hostname'])
        self.assertEqual(o.comment, d['comment'])
        self.assertEqual(o.hold_job, d['hold_job'])
        self.assertEqual(o.host_keys, d['host_keys'])
        self.assertEqual(o.connection_port, d['connection_port'])
        self.assertEqual(o.attributes, d['attributes'])

    def test_custom_connection_port(self):
        n = zk.Node('0001')
        n.state = zk.BUILDING
        d = n.toDict()
        self.assertEqual(d["connection_port"], 22, "Default port not 22")
        n = zk.Node.fromDict(d, '0001')
        self.assertEqual(n.connection_port, 22, "Default port not 22")
        n.connection_port = 22022
        d = n.toDict()
        self.assertEqual(d["connection_port"], 22022,
                         "Custom ssh port not set")



# Copyright (C) 2020 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import tempfile
import time
from unittest.mock import patch

import yaml

import googleapiclient.discovery
import googleapiclient.errors

from nodepool import tests
from nodepool import zk
from nodepool.nodeutils import iterate_timeout


class GCloudRequest:
    def __init__(self, method, args, kw):
        self.method = method
        self.args = args
        self.kw = kw

    def execute(self):
        return self.method(*self.args, **self.kw)


class GCloudCollection:
    def __init__(self):
        self.items = []

    def list(self, *args, **kw):
        return GCloudRequest(self._list, args, kw)

    def _list(self, *args, **kw):
        return dict(
            items=self.items,
        )

    def insert(self, *args, **kw):
        return GCloudRequest(self._insert, args, kw)

    def delete(self, *args, **kw):
        return GCloudRequest(self._delete, args, kw)


class GCloudInstances(GCloudCollection):
    def _insert(self, *args, **kw):
        item = kw['body'].copy()
        item['status'] = 'RUNNING'
        item['zone'] = ('https://www.googleapis.com/compute/v1/projects/'
                        + kw['project'] + '/' + kw['zone'])
        item['networkInterfaces'][0]['networkIP'] = '10.0.0.1'
        item['networkInterfaces'][0]['accessConfigs'][0]['natIP'] = '8.8.8.8'
        item['selfLink'] = ("https://www.googleapis.com/compute/v1/projects/"
                            + kw['project'] + '/instances/'
                            + kw['body']['name'])
        self.items.append(item)

    def _delete(self, *args, **kw):
        for item in self.items[:]:
            if (kw['zone'] in item['zone'] and
                kw['instance'] == item['name'] and
                kw['project'] in item['selfLink']):
                self.items.remove(item)


class GCloudImages(GCloudCollection):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.items.append({
            "family": "debian-9",
            "selfLink": "https://www.googleapis.com/compute/beta/projects/"
            "debian-cloud/global/images/debian-9-stretch-v20200309",
        })

    def getFromFamily(self, *args, **kw):
        return GCloudRequest(self._getFromFamily, args, kw)

    def _getFromFamily(self, *args, **kw):
        for item in self.items:
            if (kw['family'] == item['family'] and
                kw['project'] in item['selfLink']):
                return item
        # Note this isn't quite right, but at least it's the correct class
        raise googleapiclient.errors.HttpError(404, b'')


class GCloudComputeEmulator:
    def __init__(self):
        self._instances = GCloudInstances()
        self._images = GCloudImages()

    def instances(self):
        return self._instances

    def images(self):
        return self._images


class GCloudEmulator:
    def __init__(self):
        self.compute = GCloudComputeEmulator()

    def build(self, *args, **kw):
        return self.compute


class TestDriverGce(tests.DBTestCase):
    log = logging.getLogger("nodepool.TestDriverAws")

    def _wait_for_provider(self, nodepool, provider):
        for _ in iterate_timeout(30, Exception, 'wait for provider'):
            try:
                nodepool.getProviderManager(provider)
                break
            except Exception:
                pass

    def _test_gce_machine(self, label,
                          is_valid_config=True,
                          host_key_checking=True):
        self.patch(googleapiclient, 'discovery', GCloudEmulator())

        conf_template = os.path.join(
            os.path.dirname(__file__), '..', 'fixtures', 'gce.yaml')
        with open(conf_template) as f:
            raw_config = yaml.safe_load(f)
        raw_config['zookeeper-servers'][0] = {
            'host': self.zookeeper_host,
            'port': self.zookeeper_port,
            'chroot': self.zookeeper_chroot,
        }

        with tempfile.NamedTemporaryFile() as tf:
            tf.write(yaml.safe_dump(
                raw_config, default_flow_style=False).encode('utf-8'))
            tf.flush()
            configfile = self.setup_config(tf.name)
            pool = self.useNodepool(configfile, watermark_sleep=1)
            pool.start()

            self._wait_for_provider(pool, 'gcloud-provider')

            with patch('nodepool.driver.simple.nodescan') as nodescan:
                nodescan.return_value = 'MOCK KEY'
                req = zk.NodeRequest()
                req.state = zk.REQUESTED
                req.node_types.append(label)
                self.zk.storeNodeRequest(req)

                self.log.debug("Waiting for request %s", req.id)
                req = self.waitForNodeRequest(req)
                self.log.debug("Finished request %s", req.id)

                if is_valid_config is False:
                    self.assertEqual(req.state, zk.FAILED)
                    self.assertEqual(req.nodes, [])
                    return

                self.assertEqual(req.state, zk.FULFILLED)
                self.assertNotEqual(req.nodes, [])

                node = self.zk.getNode(req.nodes[0])
                self.assertEqual(node.allocated_to, req.id)
                self.assertEqual(node.state, zk.READY)
                self.assertIsNotNone(node.launcher)
                self.assertEqual(node.connection_type, 'ssh')
                self.assertEqual(node.attributes,
                                 {'key1': 'value1', 'key2': 'value2'})
                if host_key_checking:
                    nodescan.assert_called_with(
                        node.interface_ip,
                        port=22,
                        timeout=180,
                        gather_hostkeys=True)

                # A new request will be paused and for lack of quota
                # until this one is deleted
                req2 = zk.NodeRequest()
                req2.state = zk.REQUESTED
                req2.node_types.append(label)
                self.zk.storeNodeRequest(req2)
                req2 = self.waitForNodeRequest(
                    req2, (zk.PENDING, zk.FAILED, zk.FULFILLED))
                self.assertEqual(req2.state, zk.PENDING)
                # It could flip from PENDING to one of the others,
                # so sleep a bit and be sure
                time.sleep(1)
                req2 = self.waitForNodeRequest(
                    req2, (zk.PENDING, zk.FAILED, zk.FULFILLED))
                self.assertEqual(req2.state, zk.PENDING)

                node.state = zk.DELETING
                self.zk.storeNode(node)

                self.waitForNodeDeletion(node)

                req2 = self.waitForNodeRequest(req2,
                                               (zk.FAILED, zk.FULFILLED))
                self.assertEqual(req2.state, zk.FULFILLED)
                node = self.zk.getNode(req2.nodes[0])
                node.state = zk.DELETING
                self.zk.storeNode(node)
                self.waitForNodeDeletion(node)

    def test_gce_machine(self):
        self._test_gce_machine('debian-stretch-f1-micro')



# Copyright 2018 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from azure.common.client_factory import get_client_from_auth_file
from azure.mgmt.resource import ResourceManagementClient
from azure.mgmt.network import NetworkManagementClient
from azure.mgmt.compute import ComputeManagementClient
from msrestazure.azure_exceptions import CloudError

from nodepool.driver import Provider
from nodepool.driver.azure import handler
from nodepool import zk


class AzureProvider(Provider):
    log = logging.getLogger("nodepool.driver.azure.AzureProvider")

    API_VERSION_COMPUTE = "2019-12-01"
    API_VERSION_DISKS = "2019-11-01"
    API_VERSION_NETWORK = "2020-03-01"
    API_VERSION_RESOURCE = "2019-10-01"

    def __init__(self, provider, *args):
        self.provider = provider
        self.zuul_public_key = provider.zuul_public_key
        self.compute_client = None
        self.disks_client = None
        self.network_client = None
        self.resource_client = None
        self.resource_group = provider.resource_group
        self.resource_group_location = provider.resource_group_location
        self._zk = None

    def start(self, zk_conn):
        self.log.debug("Starting")
        self._zk = zk_conn
        self.log.debug(
            "Using %s as auth_path for Azure auth" % self.provider.auth_path)
        if self.compute_client is None:
            self.compute_client = self._get_compute_client()
        if self.disks_client is None:
            self.disks_client = self._get_disks_client()
        if self.network_client is None:
            self.network_client = self._get_network_client()
        if self.resource_client is None:
            self.resource_client = self._get_resource_client()

    def _get_compute_client(self):
        return get_client_from_auth_file(
            ComputeManagementClient,
            auth_path=self.provider.auth_path,
            api_version=self.API_VERSION_COMPUTE
        )

    def _get_disks_client(self):
        return get_client_from_auth_file(
            ComputeManagementClient,
            auth_path=self.provider.auth_path,
            api_version=self.API_VERSION_DISKS
        )

    def _get_network_client(self):
        return get_client_from_auth_file(
            NetworkManagementClient,
            auth_path=self.provider.auth_path,
            api_version=self.API_VERSION_NETWORK
        )

    def _get_resource_client(self):
        return get_client_from_auth_file(
            ResourceManagementClient,
            auth_path=self.provider.auth_path,
            api_version=self.API_VERSION_RESOURCE
        )

    def stop(self):
        self.log.debug("Stopping")

    def listNodes(self):
        return self.compute_client.virtual_machines.list(self.resource_group)

    def listNICs(self):
        return self.network_client.network_interfaces.list(self.resource_group)

    def listPIPs(self):
        return self.network_client.public_ip_addresses.list(
            self.resource_group)

    def listDisks(self):
        return self.disks_client.disks.list_by_resource_group(
            self.resource_group)

    def labelReady(self, name):
        return True

    def join(self):
        return True

    def getRequestHandler(self, poolworker, request):
        return handler.AzureNodeRequestHandler(poolworker, request)

    def cleanupLeakedResources(self):
        self._cleanupLeakedNodes()
        self._cleanupLeakedNICs()
        self._cleanupLeakedPIPs()
        self._cleanupLeakedDisks()

    def _cleanupLeakedDisks(self):
        for disk in self.listDisks():
            if disk.tags is None:
                # Nothing to check ownership against, move on
                continue
            if 'nodepool_provider_name' not in disk.tags:
                continue
            if disk.tags['nodepool_provider_name'] != self.provider.name:
                # Another launcher, sharing this provider but configured
                # with a different name, owns this.
                continue
            if not self._zk.getNode(disk.tags['nodepool_id']):
                self.log.warning(
                    "Marking for delete leaked Disk %s (%s) in %s "
                    "(unknown node id %s)",
                    disk.name, disk.id, self.provider.name,
                    disk.tags['nodepool_id']
                )
                try:
                    self.disks_client.disks.delete(
                        self.resource_group,
                        disk.name).wait()
                except CloudError as e:
                    self.log.warning(
                        "Failed to cleanup Disk %s (%s). Error: %r",
                        disk.name, disk.id, e
                    )

    def _cleanupLeakedNICs(self):
        for nic in self.listNICs():
            if nic.tags is None:
                # Nothing to check ownership against, move on
                continue
            if 'nodepool_provider_name' not in nic.tags:
                continue
            if nic.tags['nodepool_provider_name'] != self.provider.name:
                # Another launcher, sharing this provider but configured
                # with a different name, owns this.
                continue
            if not self._zk.getNode(nic.tags['nodepool_id']):
                self.log.warning(
                    "Marking for delete leaked NIC %s (%s) in %s "
                    "(unknown node id %s)",
                    nic.name, nic.id, self.provider.name,
                    nic.tags['nodepool_id']
                )
                try:
                    self.network_client.network_interfaces.delete(
                        self.resource_group,
                        nic.name).wait()
                except CloudError as e:
                    self.log.warning(
                        "Failed to cleanup NIC %s (%s). Error: %r",
                        nic.name, nic.id, e
                    )

    def _cleanupLeakedPIPs(self):
        for pip in self.listPIPs():
            if pip.tags is None:
                # Nothing to check ownership against, move on
                continue
            if 'nodepool_provider_name' not in pip.tags:
                continue
            if pip.tags['nodepool_provider_name'] != self.provider.name:
                # Another launcher, sharing this provider but configured
                # with a different name, owns this.
                continue
            if not self._zk.getNode(pip.tags['nodepool_id']):
                self.log.warning(
                    "Marking for delete leaked PIP %s (%s) in %s "
                    "(unknown node id %s)",
                    pip.name, pip.id, self.provider.name,
                    pip.tags['nodepool_id']
                )
                try:
                    self.network_client.public_ip_addresses.delete(
                        self.resource_group,
                        pip.name).wait()
                except CloudError as e:
                    self.log.warning(
                        "Failed to cleanup IP %s (%s). Error: %r",
                        pip.name, pip.id, e
                    )

    def _cleanupLeakedNodes(self):

        deleting_nodes = {}

        for node in self._zk.nodeIterator():
            if node.state == zk.DELETING:
                if node.provider != self.provider.name:
                    continue
                if node.provider not in deleting_nodes:
                    deleting_nodes[node.provider] = []
                deleting_nodes[node.provider].append(node.external_id)

        for n in self.listNodes():
            if n.tags is None:
                # Nothing to check ownership against, move on
                continue
            if 'nodepool_provider_name' not in n.tags:
                continue
            if n.tags['nodepool_provider_name'] != self.provider.name:
                # Another launcher, sharing this provider but configured
                # with a different name, owns this.
                continue
            if (self.provider.name in deleting_nodes and
                n.id in deleting_nodes[self.provider.name]):
                # Already deleting this node
                continue
            if not self._zk.getNode(n.tags['nodepool_id']):
                self.log.warning(
                    "Marking for delete leaked instance %s (%s) in %s "
                    "(unknown node id %s)",
                    n.name, n.id, self.provider.name,
                    n.tags['nodepool_id']
                )
                node = zk.Node()
                node.external_id = n.id
                node.provider = self.provider.name
                node.state = zk.DELETING
                self._zk.storeNode(node)

    def cleanupNode(self, server_id):
        self.log.debug('Server ID: %s' % server_id)
        try:
            vm = self.compute_client.virtual_machines.get(
                self.resource_group, server_id.rsplit('/', 1)[1])
        except CloudError as e:
            if e.status_code == 404:
                return
            self.log.warning(
                "Failed to cleanup node %s. Error: %r",
                server_id, e
            )

        self.compute_client.virtual_machines.delete(
            self.resource_group, server_id.rsplit('/', 1)[1]).wait()

        nic_deletion = self.network_client.network_interfaces.delete(
            self.resource_group, "%s-nic" % server_id.rsplit('/', 1)[1])
        nic_deletion.wait()

        pip_deletion = self.network_client.public_ip_addresses.delete(
            self.resource_group, "%s-nic-pip" % server_id.rsplit('/', 1)[1])
        pip_deletion.wait()

        if self.provider.ipv6:
            pip_deletion = self.network_client.public_ip_addresses.delete(
                self.resource_group,
                "%s-nic-v6-pip" % server_id.rsplit('/', 1)[1])
            pip_deletion.wait()

        disk_handle_list = []
        for disk in self.listDisks():
            if disk.tags is not None and \
                disk.tags.get('nodepool_id') == vm.tags['nodepool_id']:
                async_disk_delete = self.disks_client.disks.delete(
                    self.resource_group, disk.name)
                disk_handle_list.append(async_disk_delete)
        for async_disk_delete in disk_handle_list:
            async_disk_delete.wait()

    def waitForNodeCleanup(self, server_id):
        # All async tasks are handled in cleanupNode
        return True

    def getInstance(self, server_id):
        return self.compute_client.virtual_machines.get(
            self.resource_group, server_id, expand='instanceView')

    def createInstance(
            self, hostname, label, nodepool_id, nodepool_node_label=None):

        self.log.debug("Create resouce group")

        tags = label.tags or {}
        tags['nodepool_provider_name'] = self.provider.name
        if nodepool_node_label:
            tags['nodepool_node_label'] = nodepool_node_label

        self.resource_client.resource_groups.create_or_update(
            self.resource_group, {
                'location': self.provider.resource_group_location,
                'tags': tags
            })
        tags['nodepool_id'] = nodepool_id
        v4_params_create = {
            'location': self.provider.location,
            'public_ip_allocation_method': 'dynamic',
            'tags': tags,
        }
        v4_pip_poll = self.network_client.public_ip_addresses.create_or_update(
            self.resource_group,
            "%s-nic-pip" % hostname,
            v4_params_create,
        )
        v4_public_ip = v4_pip_poll.result()

        nic_data = {
            'location': self.provider.location,
            'tags': tags,
            'ip_configurations': [{
                'name': "zuul-v4-ip-config",
                'private_ip_address_version': 'IPv4',
                'subnet': {
                    'id': self.provider.subnet_id
                },
                'public_ip_address': {
                    'id': v4_public_ip.id
                }
            }]
        }

        if self.provider.ipv6:
            nic_data['ip_configurations'].append({
                'name': "zuul-v6-ip-config",
                'private_ip_address_version': 'IPv6',
                'subnet': {
                    'id': self.provider.subnet_id
                }
            })

        nic_creation = self.network_client.network_interfaces.create_or_update(
            self.resource_group,
            "%s-nic" % hostname,
            nic_data
        )

        nic = nic_creation.result()

        vm_creation = self.compute_client.virtual_machines.create_or_update(
            self.resource_group, hostname, {
                'location': self.provider.location,
                'os_profile': {
                    'computer_name': hostname,
                    'admin_username': label.username,
                    'linux_configuration': {
                        'ssh': {
                            'public_keys': [{
                                'path': "/home/%s/.ssh/authorized_keys" % (
                                    label.username),
                                'key_data': self.provider.zuul_public_key,
                            }]
                        },
                        "disable_password_authentication": True,
                    }
                },
                'hardware_profile': {
                    'vmSize': label.hardwareProfile["vm-size"]
                },
                'storage_profile': {'image_reference': label.imageReference},
                'network_profile': {
                    'network_interfaces': [{
                        'id': nic.id,
                        'properties': {
                            'primary': True,
                        }
                    }]
                },
                'tags': tags,
            })
        return vm_creation.result()

    def getIpaddress(self, instance):
        # Copied from https://github.com/Azure/azure-sdk-for-python/issues/897
        ni_reference = instance.network_profile.network_interfaces[0]
        ni_reference = ni_reference.id.split('/')
        ni_group = ni_reference[4]
        ni_name = ni_reference[8]

        net_interface = self.network_client.network_interfaces.get(
            ni_group, ni_name)
        ip_reference = net_interface.ip_configurations[0].public_ip_address
        ip_reference = ip_reference.id.split('/')
        ip_group = ip_reference[4]
        ip_name = ip_reference[8]

        public_ip = self.network_client.public_ip_addresses.get(
            ip_group, ip_name)
        public_ip = public_ip.ip_address
        return public_ip

    def getv6Ipaddress(self, instance):
        # Copied from https://github.com/Azure/azure-sdk-for-python/issues/897
        ni_reference = instance.network_profile.network_interfaces[0]
        ni_reference = ni_reference.id.split('/')
        ni_group = ni_reference[4]
        ni_name = ni_reference[8]

        net_interface = self.network_client.network_interfaces.get(
            ni_group, ni_name)
        return net_interface.ip_configurations[1].private_ip_address



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import signal
import sys

from nodepool import builder
import nodepool.cmd


class NodePoolBuilderApp(nodepool.cmd.NodepoolDaemonApp):

    app_name = 'nodepool-builder'
    app_description = 'NodePool Image Builder.'

    def sigint_handler(self, signal, frame):
        self.nb.stop()
        sys.exit(0)

    def create_parser(self):
        parser = super(NodePoolBuilderApp, self).create_parser()

        parser.add_argument('-c', dest='config',
                            default='/etc/nodepool/nodepool.yaml',
                            help='path to config file')
        parser.add_argument('-s', dest='secure',
                            help='path to secure config file')
        parser.add_argument('--build-workers', dest='build_workers',
                            default=1, help='number of build workers',
                            type=int)
        parser.add_argument('--upload-workers', dest='upload_workers',
                            default=4, help='number of upload workers',
                            type=int)
        return parser

    def parse_args(self):
        super(NodePoolBuilderApp, self).parse_args()
        self.config_file = self.get_path(self.args.config)
        self.secure_file = self.get_path(self.args.secure)

    def run(self):
        self.nb = builder.NodePoolBuilder(
            self.config_file,
            secure_path=self.secure_file,
            num_builders=self.args.build_workers,
            num_uploaders=self.args.upload_workers)

        signal.signal(signal.SIGINT, self.sigint_handler)

        self.nb.start()

        while True:
            signal.pause()


def main():
    return NodePoolBuilderApp.main()



# Copyright 2019 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from nodepool.driver.simple import SimpleTaskManagerAdapter
from nodepool.driver.simple import SimpleTaskManagerInstance

import googleapiclient.discovery


class GCEInstance(SimpleTaskManagerInstance):
    def load(self, data):
        if data['status'] == 'TERMINATED':
            self.deleted = True
        elif data['status'] == 'RUNNING':
            self.ready = True
        self.external_id = data['name']
        self.az = data['zone']

        iface = data.get('networkInterfaces', [])
        if len(iface):
            self.private_ipv4 = iface[0].get('networkIP')
            access = iface[0].get('accessConfigs', [])
            if len(access):
                self.public_ipv4 = access[0].get('natIP')
        self.interface_ip = self.public_ipv4 or self.private_ipv4

        if data.get('metadata'):
            for item in data['metadata'].get('items', []):
                self.metadata[item['key']] = item['value']


class GCEAdapter(SimpleTaskManagerAdapter):
    log = logging.getLogger("nodepool.driver.gce.GCEAdapter")

    def __init__(self, provider):
        self.provider = provider
        self.compute = googleapiclient.discovery.build('compute', 'v1')

    def listInstances(self, task_manager):
        servers = []

        q = self.compute.instances().list(project=self.provider.project,
                                          zone=self.provider.zone)
        with task_manager.rateLimit():
            result = q.execute()

        for instance in result.get('items', []):
            servers.append(GCEInstance(instance))
        return servers

    def deleteInstance(self, task_manager, server_id):
        q = self.compute.instances().delete(project=self.provider.project,
                                            zone=self.provider.zone,
                                            instance=server_id)
        with task_manager.rateLimit():
            q.execute()

    def _getImageId(self, task_manager, cloud_image):
        image_id = cloud_image.image_id

        if image_id:
            return image_id

        if cloud_image.image_family:
            q = self.compute.images().getFromFamily(
                project=cloud_image.image_project,
                family=cloud_image.image_family)
            with task_manager.rateLimit():
                result = q.execute()
            image_id = result['selfLink']

        return image_id

    def createInstance(self, task_manager, hostname, metadata, label):
        image_id = self._getImageId(task_manager, label.cloud_image)
        disk_init = dict(sourceImage=image_id,
                         diskType='zones/{}/diskTypes/{}'.format(
                             self.provider.zone, label.volume_type),
                         diskSizeGb=str(label.volume_size))
        disk = dict(boot=True,
                    autoDelete=True,
                    initializeParams=disk_init)
        machine_type = 'zones/{}/machineTypes/{}'.format(
            self.provider.zone, label.instance_type)
        network = dict(network='global/networks/default',
                       accessConfigs=[dict(
                           type='ONE_TO_ONE_NAT',
                           name='External NAT')])
        metadata_items = []
        for (k, v) in metadata.items():
            metadata_items.append(dict(key=k, value=v))
        meta = dict(items=metadata_items)
        args = dict(
            name=hostname,
            machineType=machine_type,
            disks=[disk],
            networkInterfaces=[network],
            serviceAccounts=[],
            metadata=meta)
        q = self.compute.instances().insert(
            project=self.provider.project,
            zone=self.provider.zone,
            body=args)
        with task_manager.rateLimit():
            q.execute()
        return hostname



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from nodepool import tests
from nodepool.config import Config
from nodepool.config import DiskImage
from nodepool.config import Label
from nodepool.driver import ConfigPool
from nodepool.driver import DriverConfig
from nodepool.driver.openstack.config import OpenStackProviderConfig
from nodepool.driver.openstack.config import ProviderDiskImage
from nodepool.driver.openstack.config import ProviderCloudImage
from nodepool.driver.openstack.config import ProviderLabel
from nodepool.driver.openstack.config import ProviderPool
from nodepool.driver.static.config import StaticPool
from nodepool.driver.static.config import StaticProviderConfig


class TempConfigPool(ConfigPool):
    def load(self):
        pass


class TestConfigComparisons(tests.BaseTestCase):

    def test_ConfigPool(self):

        a = TempConfigPool()
        b = TempConfigPool()
        self.assertEqual(a, b)
        a.max_servers = 5
        self.assertNotEqual(a, b)

    def test_DriverConfig(self):
        a = DriverConfig()
        b = DriverConfig()
        self.assertEqual(a, b)
        a.name = "foo"
        self.assertNotEqual(a, b)

    def test_Config(self):
        a = Config()
        b = Config()
        self.assertEqual(a, b)
        a.imagesdir = "foo"
        self.assertNotEqual(a, b)

    def test_Label(self):
        a = Label()
        b = Label()
        self.assertEqual(a, b)
        a.name = "foo"
        self.assertNotEqual(a, b)

    def test_DiskImage(self):
        a = DiskImage('foo')
        b = DiskImage('foo')
        self.assertEqual(a, b)
        a.name = 'bar'
        self.assertNotEqual(a, b)

    def test_ProviderDiskImage(self):
        a = ProviderDiskImage()
        b = ProviderDiskImage()
        self.assertEqual(a, b)
        a.name = "foo"
        self.assertNotEqual(a, b)

    def test_ProviderCloudImage(self):
        a = ProviderCloudImage()
        b = ProviderCloudImage()
        self.assertEqual(a, b)
        a.name = "foo"
        self.assertNotEqual(a, b)

    def test_ProviderLabel(self):
        a = ProviderLabel()
        b = ProviderLabel()
        self.assertEqual(a, b)
        a.name = "foo"
        self.assertNotEqual(a, b)

    def test_ProviderPool(self):
        a = ProviderPool()
        b = ProviderPool()
        self.assertEqual(a, b)
        # intentionally change an attribute of the base class
        a.max_servers = 5
        self.assertNotEqual(a, b)

        c = TempConfigPool()
        d = ProviderPool()
        self.assertNotEqual(d, c)

    def test_OpenStackProviderConfig(self):
        provider = {'name': 'foo'}
        a = OpenStackProviderConfig(None, provider)
        b = OpenStackProviderConfig(None, provider)
        self.assertEqual(a, b)
        # intentionally change an attribute of the base class
        a.name = 'bar'
        self.assertNotEqual(a, b)

    def test_StaticPool(self):
        a = StaticPool()
        b = StaticPool()
        self.assertEqual(a, b)
        # intentionally change an attribute of the base class
        a.max_servers = 5
        self.assertNotEqual(a, b)
        c = TempConfigPool()
        self.assertNotEqual(b, c)

    def test_StaticProviderConfig(self):
        provider = {'name': 'foo'}
        a = StaticProviderConfig(provider)
        b = StaticProviderConfig(provider)
        self.assertEqual(a, b)
        # intentionally change an attribute of the base class
        a.name = 'bar'
        self.assertNotEqual(a, b)
